import pytest
import agentpy as ap

from agentpy.tools import AgentpyError


def test_make_list():

    make_list = ap.tools.make_list
    assert make_list('123') == ['123']
    assert make_list(['123']) == ['123']
    assert make_list(None) == []
    assert make_list(None, keep_none=True) == [None]


def test_attr_dict():

    ad = ap.AttrDict({'a': 1})
    ad.b = 2

    assert ad.a == 1
    assert ad.b == 2
    assert ad.a == ad['a']
    assert ad.b == ad['b']
    assert ad.__repr__() == "AttrDict {'a': 1, 'b': 2}"
    assert ad._short_repr() == "AttrDict {2 entries}"



"""
Agentpy Experiment Module
Content: Experiment class
"""

import pandas as pd
import ipywidgets
import IPython

from datetime import datetime, timedelta
from .tools import make_list
from .output import DataDict


class Experiment:
    """ Experiment for an agent-based model.
    Allows for multiple iterations, parameter samples, scenario comparison,
    and parallel processing. See :func:`Experiment.run` for standard
    simulations and :func:`Experiment.interactive` for interactive output.

    Arguments:
        model_class(type): The model class type that the experiment should use.
        parameters(dict or list of dict, optional):
            Parameter dictionary or sample (default None).
        name(str, optional): Name of the experiment (default model.name).
        scenarios(str or list, optional): Experiment scenarios (default None).
        iterations(int, optional): Experiment repetitions (default 1).
        record(bool, optional):
            Whether to keep the record of dynamic variables (default False).
            Note that this does not affect evaluation measures.
        **kwargs: Will be forwarded to the creation of every model instance
            during the experiment.

    Attributes:
        output(DataDict): Recorded experiment data
    """

    def __init__(self, model_class, parameters=None, name=None, scenarios=None,
                 iterations=1, record=False, **kwargs):

        self.model = model_class
        self.output = DataDict()
        self.iterations = iterations
        self.record = record
        self._model_kwargs = kwargs

        if name:
            self.name = name
        else:
            self.name = model_class.__name__

        # Transform input into iterable lists if only a single value is given
        # keep_none assures that make_list(None) returns iterable [None]
        self.scenarios = make_list(scenarios, keep_none=True)
        self.parameters = make_list(parameters, keep_none=True)
        self._parameters_to_output()  # Record parameters

        # Log
        self.output.log = {'name': self.name,
                           'model_type': model_class.__name__,
                           'time_stamp': str(datetime.now()),
                           'iterations': iterations}
        if scenarios:
            self.output.log['scenarios'] = scenarios

        # Prepare runs
        self.parameters_per_run = self.parameters * self.iterations
        self.number_of_runs = len(self.parameters_per_run)

    def _parameters_to_output(self):
        """ Document parameters (seperately for fixed & variable). """
        df = pd.DataFrame(self.parameters)
        df.index.rename('sample_id', inplace=True)
        fixed_pars = {}
        for col in df.columns:
            s = df[col]
            if len(s.unique()) == 1:
                fixed_pars[s.name] = df[col][0]
                df.drop(col, inplace=True, axis=1)
        if fixed_pars and df.empty:
            self.output['parameters'] = fixed_pars
        elif not fixed_pars and not df.empty:
            self.output['parameters'] = df
        else:
            self.output['parameters'] = DataDict({
                'fixed': fixed_pars,
                'varied': df
            })

    def _add_single_output_to_combined(self, single_output, combined_output):
        """Append results from single run to combined output.
        Each key in single_output becomes a key in combined_output.
        DataDicts entries become dicts with lists of values.
        Other entries become lists of values. """
        for key, value in single_output.items():
            if key in ['parameters', 'log']:  # Skip parameters & log
                continue
            if isinstance(value, DataDict):  # Handle subdicts
                if key not in combined_output:  # New key
                    combined_output[key] = {}  # as dict
                for obj_type, obj_df in single_output[key].items():
                    if obj_type not in combined_output[key]:  # New subkey
                        combined_output[key][obj_type] = []  # as list
                    combined_output[key][obj_type].append(obj_df)
            else:  # Handle other output types
                if key not in combined_output:  # New key
                    combined_output[key] = []  # as list
                combined_output[key].append(value)

    def _combine_dataframes(self, combined_output):
        """ Combines data from combined output.
        Dataframes are combined with concat.
        Dicts are transformed to DataDict.
        Other objects are kept as original.
        Combined data is written to self.output. """
        for key, values in combined_output.items():
            if values and all([isinstance(value, pd.DataFrame)
                               for value in values]):
                self.output[key] = pd.concat(values)  # Df are combined
            elif isinstance(values, dict):  # Dict is transformed to DataDict
                self.output[key] = DataDict()
                for sk, sv in values.items():
                    if all([isinstance(v, pd.DataFrame) for v in sv]):
                        self.output[key][sk] = pd.concat(sv)  # Df are combined
                    else:  # Other objects are kept as original TODO TESTS
                        self.output[key][sk] = sv
            elif key != 'log':  # Other objects are kept as original TODO TESTS
                self.output[key] = values

    def _single_sim(self, sim_id):
        """ Perform a single simulation."""
        sc_id = sim_id % len(self.scenarios)
        run_id = (sim_id - sc_id) // len(self.scenarios)
        model = self.model(
            self.parameters_per_run[run_id],
            run_id=run_id,
            scenario=self.scenarios[sc_id],
            **self._model_kwargs)
        results = model.run(display=False)
        if 'variables' in results and self.record is False:
            del results['variables']  # Remove dynamic variables from record
        return results

    def run(self, pool=None, display=True):
        """ Executes a multi-run experiment.

        The simulation will run the model once for each set of parameters
        and will repeat this process for the set number of iterations.
        Parallel processing is possible if a `pool` is passed.
        Simulation results will be stored in `Experiment.output`.

        Arguments:
            pool(multiprocessing.Pool, optional):
                Pool of active processes for parallel processing.
                If none is passed, normal processing is used.
            display(bool, optional):
                Display simulation progress (default True).

        Returns:
            DataDict: Recorded experiment data.

        Examples:

            To run a normal experiment::

                exp = ap.Experiment(MyModel, parameters)
                results = exp.run()

            To use parallel processing::

                import multiprocessing as mp
                if __name__ ==  '__main__':
                    exp = ap.Experiment(MyModel, parameters)
                    pool = mp.Pool(mp.cpu_count())
                    results = exp.run(pool)
        """

        sim_ids = list(range(self.number_of_runs * len(self.scenarios)))
        n_sims = len(sim_ids)
        if display:
            print(f"Scheduled runs: {n_sims}")
        t0 = datetime.now()  # Time-Stamp Start
        combined_output = {}

        if pool is None:  # Normal processing
            for sim_id in sim_ids:
                self._add_single_output_to_combined(
                    self._single_sim(sim_id), combined_output)
                if display:
                    td = (datetime.now() - t0).total_seconds()
                    te = timedelta(seconds=int(td / (sim_id + 1)
                                               * (n_sims - sim_id - 1)))
                    print(f"\rCompleted: {sim_id + 1}, "
                          f"estimated time remaining: {te}", end='')
            if display:
                print("")  # Because the last print ended without a line-break
        else:  # Parallel processing
            if display:
                print(f"Active processes: {pool._processes}")
            output_list = pool.map(self._single_sim, sim_ids)
            for single_output in output_list:
                self._add_single_output_to_combined(
                    single_output, combined_output)

        self._combine_dataframes(combined_output)
        self.output.log['run_time'] = ct = str(datetime.now() - t0)

        if display:
            print(f"Experiment finished\nRun time: {ct}")

        return self.output

    def interactive(self, plot, *args, **kwargs):
        """
        Displays interactive output for Jupyter notebooks,
        using :mod:`IPython` and :mod:`ipywidgets`.
        A slider will be shown for varied parameters.
        Every time a parameter value is changed on the slider,
        the experiment will re-run the model and pass it
        to the 'plot' function.

        Arguments:
            plot: Function that takes a model instance as input
                and prints or plots the desired output..
            *args: Will be forwarded to 'plot'.
            **kwargs: Will be forwarded to 'plot'.

        Returns:
            ipywidgets.HBox: Interactive output widget
            
        Examples:
            The following example uses a custom model :class:`MyModel`
            and creates a slider for the parameters 'x' and 'y',
            both of which can be varied interactively over 10 different values.
            Every time a value is changed, the experiment will simulate the
            model with the new parameters and pass it to the plot function::
            
                def plot(model):
                    # Display interactive output here
                    print(model.output)
                    
                param_ranges = {'x': (0, 10), 'y': (0., 1.)}
                sample = ap.sample(param_ranges, n=10)
                exp = ap.Experiment(MyModel, sample)
                exp.interactive(plot)
        """

        def var_run(**param_updates):
            """ Display plot for updated parameters. """
            IPython.display.clear_output()
            parameters = dict(self.parameters[0])
            parameters.update(param_updates)
            temp_model = self.model(parameters, **self._model_kwargs)
            temp_model.run()
            IPython.display.clear_output()
            plot(temp_model, *args, **kwargs)

        # Get variable parameters
        var_pars = self.output._combine_pars(varied=True, fixed=False)

        # Create widget dict
        widget_dict = {}
        for par_key in list(var_pars):
            par_list = list(var_pars[par_key])

            widget_dict[par_key] = ipywidgets.SelectionSlider(
                options=par_list,
                value=par_list[0],
                description=par_key,
                continuous_update=False,
                style=dict(description_width='initial'),
                layout={'width': '300px'}
            )

        widgets_left = ipywidgets.VBox(list(widget_dict.values()))
        output_right = ipywidgets.interactive_output(var_run, widget_dict)

        return ipywidgets.HBox([widgets_left, output_right])



"""
Agentpy Lists Module
Content: Lists for objects, environments, and agents
"""

import numpy as np


class AttrList(list):
    """ List of attributes from an :class:`AgentList`.

    Calls are forwarded to each entry and return a list of return values.
    Boolean operators are applied to each entry and return a list of bools.
    Arithmetic operators are applied to each entry and return a new list.
    See :class:`AgentList` for examples.
    """

    def __init__(self, iterable=[], attr=None):
        super().__init__(iterable)
        self.attr = attr

    def __repr__(self):
        if self.attr is None:
            return f"AttrList: {list.__repr__(self)}"
        else:
            return f"AttrList of attribute '{self.attr}': " \
                   f"{list.__repr__(self)}"

    def __call__(self, *args, **kwargs):
        return AttrList(
            [func_obj(*args, **kwargs) for func_obj in self],
            attr=self.attr)  # TODO Add return values to attr name?

    def __eq__(self, other):
        return [obj == other for obj in self]

    def __ne__(self, other):
        return [obj != other for obj in self]

    def __lt__(self, other):
        return [obj < other for obj in self]

    def __le__(self, other):
        return [obj <= other for obj in self]

    def __gt__(self, other):
        return [obj > other for obj in self]

    def __ge__(self, other):
        return [obj >= other for obj in self]

    def __add__(self, v):
        if isinstance(v, AttrList):
            return AttrList([x + y for x, y in zip(self, v)])
        else:
            return AttrList([x + v for x in self])

    def __sub__(self, v):
        if isinstance(v, AttrList):
            return AttrList([x - y for x, y in zip(self, v)])
        else:
            return AttrList([x - v for x in self])

    def __mul__(self, v):
        if isinstance(v, AttrList):
            return AttrList([x * y for x, y in zip(self, v)])
        else:
            return AttrList([x * v for x in self])

    def __truediv__(self, v):
        if isinstance(v, AttrList):
            return AttrList([x / y for x, y in zip(self, v)])
        else:
            return AttrList([x / v for x in self])

    def __iadd__(self, v):
        return self + v

    def __isub__(self, v):
        return self - v

    def __imul__(self, v):
        return self * v

    def __itruediv__(self, v):
        return self / v


class ObjList(list):
    """ List of agentpy objects (models, environments, agents). """

    def __init__(self, iterable=[], model=None):
        super().__init__(iterable)
        super().__setattr__('model', model)

    def __repr__(self):
        s = 's' if len(self) > 1 else ''
        return f"ObjList [{len(self)} object{s}]"

    def __setattr__(self, name, value):
        if isinstance(value, AttrList):
            # Apply each value to each agent
            for obj, v in zip(self, value):
                setattr(obj, name, v)
        else:
            # Apply single value to all agents
            for obj in self:
                setattr(obj, name, value)

    def __getattr__(self, name):
        """ Return callable list of attributes """
        return AttrList([getattr(obj, name) for obj in self], attr=name)

    def __call__(self, selection):
        return self.select(selection)

    def _default_generator(self):
        """ Try to find default number generator. """
        if self.model:
            return self.model.random
        elif len(self) > 0 and hasattr(self[0], 'model'):
            return self[0].model.random
        else:
            return np.random.default_rng()

    def select(self, selection):
        """ Returns a new :class:`AgentList` based on `selection`.

        Arguments:
            selection (list of bool): List with same length as the agent list.
                Positions that return True will be selected.
        """
        return AgentList([a for a, s in zip(self, selection) if s],
                         model=self.model)

    def random(self, n=1, replace=False, weights=None, shuffle=True,
               generator=None):
        """ Creates a random sample of agents,
        using :func:`numpy.random.Generator.choice`.
        Argument descriptions are adapted from :obj:`numpy.random`.
        Returns a new :class:`AgentList` with the selected agents.

        Arguments:
            n (int, optional): Number of agents (default 1).
            replace (bool, optional):
                Whether the sample is with or without replacement.
                Default is False, meaning that every agent can
                only be selected once.
            weights (1-D array_like, optional):
                The probabilities associated with each agent.
                If not given the sample assumes a uniform distribution
                over all agents.
            shuffle (bool, optional):
                Whether the sample is shuffled
                when sampling without replacement.
                Default is True, False provides a speedup.
            generator (numpy.random.Generator, optional):
                Random number generator.
                If none is passed, :obj:`Model.random` is used.
                If list has no model, :obj:`np.random` is used.
        """
        if not generator:
            generator = self._default_generator()
        # Choice is not applied to list directly because it would convert it to
        # a numpy array, which takes much more time than the current solution.
        indexes = generator.choice(len(self), size=n, replace=replace,
                                   p=weights, shuffle=shuffle)
        selection = AgentList([self[i] for i in indexes], model=self.model)
        return selection

    def sort(self, var_key, reverse=False):
        """ Sorts the list using :func:`list.sort`, and returns self.

        Arguments:
            var_key (str): Attribute of the lists' objects, based on which
                the list will be sorted from lowest value to highest.
            reverse (bool, optional): Reverse sorting (default False).
        """
        super().sort(key=lambda x: x[var_key], reverse=reverse)
        return self

    def shuffle(self, generator=None):
        """ Shuffles the list randomly, using :func:`generator.shuffle`.
        Returns self.

        Arguments:
            generator (numpy.random.Generator, optional):
                Random number generator.
                If none is passed, :obj:`Model.random` is used.
                If list has no model, :obj:`np.random` is used.
        """
        if not generator:
            generator = self._default_generator()
        generator.shuffle(self)
        return self


class AgentList(ObjList):
    """ List of agents.

    Attribute calls and assignments are applied to all agents
    and return an :class:`AttrList` with attributes of each agent.
    This also works for method calls, which returns a list of return values.
    Arithmetic operators can further be used to manipulate agent attributes,
    and boolean operators can be used to filter list based on agent attributes.

    Examples:

        Prepare an :class:`AgentList` with three agents::

            >>> model = ap.Model()
            >>> agents = model.add_agents(3)
            >>> agents
            AgentList [3 agents]

        The assignment operator can be used to set a variable for each agent.
        When the variable is called, an :class:`AttrList` is returned::

            >>> agents.x = 1
            >>> agents.x
            AttrList of attribute 'x': [1, 1, 1]

        One can also set different variables for each agent
        by passing another :class:`AttrList`::

            >>> agents.y = ap.AttrList([1, 2, 3])
            >>> agents.y
            AttrList of attribute 'y': [1, 2, 3]

        Arithmetic operators can be used in a similar way.
        If an :class:`AttrList` is passed, different values are used for
        each agent. Otherwise, the same value is used for all agents::

            >>> agents.x = agents.x + agents.y
            >>> agents.x
            AttrList of attribute 'x': [2, 3, 4]

            >>> agents.x *= 2
            >>> agents.x
            AttrList of attribute 'x': [4, 6, 8]

        Boolean operators can be used to select a subset of agents::

            >>> subset = agents(agents.x > 5)
            >>> subset
            AgentList [2 agents]

            >>> subset.x
            AttrList of attribute 'x': [6, 8]
    """

    def __repr__(self):
        return f"AgentList [{len(self)} agent{'s' if len(self) != 1 else ''}]"


class EnvList(ObjList):
    """ List of environments.

    Attribute calls and assignments are applied to all environments
    and return an :class:`AttrList` with attributes of each environment.
    This also works for method calls, which returns a list of return values.
    Arithmetic operators can further be used to manipulate attributes,
    and boolean operators can be used to filter list based on attributes.

    See :class:`AgentList` for examples.
    """

    def __repr__(self):
        s = 's' if len(self) > 1 else ''
        return f"EnvList [{len(self)} environment{s}]"

    def add_agents(self, *args, **kwargs):
        """ Add the same agents to all environments in the list.
        See :func:`Environment.add_agents` for arguments and keywords."""

        if self:
            new_agents = self[0].add_agents(*args, **kwargs)
            if len(self) > 1:
                for env in self[1:]:
                    env.add_agents(new_agents)



import pytest
import agentpy as ap
import numpy as np
import scipy


def test_general():

    model = ap.Model()
    space = model.add_space((2, 2))
    assert space is model.env
    assert space.shape == (2, 2)
    assert space.dim == 2


def test_KDTree():

    model = ap.Model()
    space = model.add_space((2, 2))
    assert space.KDTree is None
    assert len(space.get_agents((1, 1), 2)) == 0
    space.add_agents(1)
    assert isinstance(space.KDTree, scipy.spatial.cKDTree)
    assert len(space.get_agents((1, 1), 2)) == 1


def test_add_agents_random():
    model = ap.Model()
    model.run(steps=0, seed=1, display=False)
    space = model.add_space((2, 2))
    agent = space.add_agents(1, random=True)[0]
    assert list(agent.position()) == [1.0236432494005134, 1.9009273926518706]


def test_remove():
    model = ap.Model()
    space = model.add_space((2, 2))
    space.add_agents(2)
    agent = model.agents[0]
    space.remove_agents(agent)
    assert len(space._agent_dict) == 1
    assert len(space.agents) == 1
    assert space.agents[0].id == 3


def test_positions():
    # Disconnected space
    model = ap.Model()
    space = model.add_space((2, 2))
    space.add_agents(1)
    space.add_agents(1, positions=[(1, 2)])

    # Position reference
    a1 = model.agents[0]
    a2 = model.agents[1]
    assert list(a1.position()) == list(space.position(a1))
    assert list(a1.position()) == list(space.position(a1.id))
    assert list(a1.position()) == [0, 0]
    assert list(a2.position()) == [1, 2]
    assert [list(x) for x in space.positions()] == [[0, 0], [1, 2]]
    assert list(space.positions(transpose=True)) == [[0, 1], [0, 2]]

    # Get agents
    assert len(space.get_agents((1, 1), 0.9)) == 0
    assert len(space.get_agents((1, 1), 1)) == 1
    assert len(space.get_agents((1, 1), 1.4)) == 1
    assert len(space.get_agents((1, 1), 1.42)) == 2

    # Get neighbors
    assert len(a1.neighbors(distance=2.0)) == 0
    assert len(a1.neighbors(distance=2.5)) == 1
    assert a1.neighbors(distance=2.5)[0] is a2

    # Movement restricted by border
    a2.move_by((2, -3))
    assert list(a2.position()) == [2, 0]

    # Move directly
    space.move_agent(a2.id, (1, 1))
    assert list(a2.position()) == [1, 1]

    # Connected space
    model = ap.Model()
    space = model.add_space((2, 2), connect_borders=True)
    space.add_agents(1)
    space.add_agents(1, positions=[(0, 1.9)])
    a1 = model.agents[0]
    a2 = model.agents[1]
    assert a1.neighbors(distance=0.11)[0] == a2

    # Movement over border
    a2.move_by((-3, 1.1))
    assert list(a2.position()) == [1, 1]



import pytest
import agentpy as ap
import pandas as pd
import shutil
import os
import multiprocessing as mp

from agentpy.tools import AgentpyError


class MyModel(ap.Model):

    def setup(self):
        self.measure('measured_id', self.model.run_id)
        self.record('t0', self.t)


def test_basics():

    exp = ap.Experiment(MyModel, [{'steps': 1}] * 3)
    results = exp.run()
    assert 'variables' not in results
    assert exp.name == 'MyModel'

    exp = ap.Experiment(MyModel, [{'steps': 1}] * 3, name='test')
    assert exp.name == 'test'

    exp = ap.Experiment(MyModel, [{'steps': 1}] * 3, record=True)
    results = exp.run()
    assert 'variables' in results


def test_parallel_processing():

    exp = ap.Experiment(MyModel, [{'steps': 1}] * 3)
    pool = mp.Pool(mp.cpu_count())
    results = exp.run(pool)

    exp2 = ap.Experiment(MyModel, [{'steps': 1}] * 3)
    results2 = exp2.run()

    del results.log
    del results2.log

    assert results == results2


def test_interactive():
    """Test only for errors."""
    def interactive_plot(m):
        print("x =", m.p.x)
    param_ranges = {'x': (0., 1.)}
    sample = ap.sample(param_ranges, n=10)
    exp = ap.Experiment(ap.Model, sample)
    exp.interactive(interactive_plot)
    assert True



import pytest
import agentpy as ap
from agentpy.tools import AgentpyError


def make_model():
    """ Create three agents (ID 1-3) in an environment (ID 4). """
    model = ap.Model()
    agents = model.add_agents(3)
    env = model.add_env(color='green')
    env.add_agents(agents)
    return model


def test_add_env():
    """ Add environment to model. """
    model = make_model()
    assert len(model.envs) == 1
    assert model.env.id == 4  # Environment is added after agents (ID 1-3)
    assert model.env.color == 'green'  # Attribute passed as kwarg
    assert type(model.env) == ap.Environment
    assert model.env == model.envs[0]
    assert model.agents == model.env.agents
    assert model.agents[0].envs == model.envs


def test_remove_agents():
    """ Remove/delete agents from environment/model. """
    model = make_model()  # Env with three agents
    assert len(model.env.agents) == 3
    model.env.remove_agents(model.agents)  # Remove from env
    assert len(model.env.agents) == 0
    assert len(model.agents) == 3
    model.remove_agents(model.agents)  # Remove from model
    assert len(model.agents) == 0
    model = make_model()
    model.agents.delete()  # Remove from everything
    assert len(model.env.agents) == 0
    assert len(model.agents) == 0


def test_enter_exit():
    """ Move agent in and out of environment. """
    model = ap.Model()
    model.add_agents()
    agent = model.get_obj(1)
    env = model.add_env()
    assert len(env.agents) == 0
    agent.enter(env)  # Call env by instance
    assert len(env.agents) == 1
    agent.exit(env)
    assert len(env.agents) == 0
    agent.enter(2)  # Call env by id
    assert len(env.agents) == 1
    agent.exit(2)
    assert len(env.agents) == 0
    agent.enter(2)
    assert len(env.agents) == 1
    agent.exit()  # Take first env by default
    assert len(env.agents) == 0
    with pytest.raises(AgentpyError):
        agent.exit()  # Agent is not part of any environment
    with pytest.raises(AgentpyError):
        agent.exit(2)  # Agent is not part of this environment


def test_neighbors_error():
    """ Neighbors cannot be called in default environment. """
    model = ap.Model()
    env = model.add_env()
    agent = model.add_agents()[0]
    with pytest.raises(AgentpyError):
        agent.neighbors()  # Agent has no suitable environment
    with pytest.raises(AgentpyError):
        agent.neighbors(env)  # Chosen environment is not suitable



"""
Agentpy Tools Module
Content: Errors, generators, and base classes
"""

from numpy import ndarray


class AgentpyError(Exception):
    pass


def make_matrix(shape, class_):
    """ Returns a nested list with given shape and class instance. """
    # H/T Thierry Lathuille https://stackoverflow.com/a/64467230/
    if len(shape) == 1:
        return [class_() for _ in range(shape[0])]
    return [make_matrix(shape[1:], class_) for _ in range(shape[0])]


def make_list(element, keep_none=False):
    """ Turns element into a list of itself
    if it is not of type list or tuple. """

    if element is None and not keep_none:
        element = []  # Convert none to empty list
    if not isinstance(element, (list, tuple, ndarray)):
        element = [element]
    elif isinstance(element, tuple):
        element = list(element)

    return element


def param_tuples_to_salib(param_ranges_tuples):
    """ Convert param_ranges to SALib Format """

    param_ranges_salib = {
        'num_vars': len(param_ranges_tuples),
        'names': list(param_ranges_tuples.keys()),
        'bounds': []
    }

    for var_key, var_range in param_ranges_tuples.items():
        param_ranges_salib['bounds'].append([var_range[0], var_range[1]])

    return param_ranges_salib


class AttrDict(dict):
    """ Dictionary where attribute calls are handled like item calls.

    Examples:

        >>> ad = ap.AttrDict()
        >>> ad['a'] = 1
        >>> ad.a
        1

        >>> ad.b = 2
        >>> ad['b']
        2
    """

    def __init__(self, *args, **kwargs):
        if args == (None, ):
            args = ()  # Empty tuple
        super().__init__(*args, **kwargs)

    def __getattr__(self, name):
        try:
            return self.__getitem__(name)
        except KeyError:
            # Important for pickle to work
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self.__setitem__(name, value)

    def __delattr__(self, item):
        del self[item]

    def __repr__(self):
        return f"AttrDict {super().__repr__()}"

    def _short_repr(self):
        len_ = len(self.keys())
        return f"AttrDict {{{len_} entr{'y' if len_ == 1 else 'ies'}}}"



import pytest
import agentpy as ap
import numpy as np
import pandas as pd
import shutil
import os

from agentpy.tools import AgentpyError


def test_combine_vars():

    model = ap.Model()
    model.record('test', 1)
    results = model.run(1, display=False)
    assert results._combine_vars().shape == (1, 1)

    model = ap.Model()
    agents = model.add_agents()
    agents.record('test', 1)
    results = model.run(1, display=False)
    assert results._combine_vars().shape == (1, 1)

    model = ap.Model()
    agents = model.add_agents()
    model.record('test', 1)
    agents.record('test', 2)
    results = model.run(1, display=False)
    assert results._combine_vars().shape == (2, 1)

    model = ap.Model()
    agents = model.add_agents()
    model.record('test', 1)
    agents.record('test', 2)
    results = model.run(1, display=False)
    assert results._combine_vars(obj_types="Model").shape == (1, 1)

    model = ap.Model()
    agents = model.add_agents()
    model.record('test', 1)
    agents.record('test', 2)
    results = model.run(1, display=False)
    assert results._combine_vars(obj_types="Doesn't exist") is None

    model = ap.Model()
    results = model.run(1, display=False)
    assert results._combine_vars() is None
    assert results._combine_pars() is None

    model = ap.Model({'test': 1})
    results = model.run(1, display=False)
    assert results._combine_pars(fixed=False) is None

    results.variables = 1
    with pytest.raises(TypeError):
        assert results._combine_vars()
    results.parameters = 1
    with pytest.raises(TypeError):
        assert results._combine_pars()


def test_arrange_scenarios():

    class MyModel(ap.Model):
        def step(self):
            self.record('test', 1)
            self.measure('test', 1)
            self.stop()

    exp = ap.Experiment(MyModel, scenarios=['sc1', 'sc2'], record=True)
    results = exp.run()

    assert results.arrange() is None
    assert results.arrange(variables='all', scenarios='sc1').shape == (1, 4)
    assert results.arrange(measures='all', scenarios='sc1').shape == (1, 3)


repr = """DataDict {
'parameters': 
    'fixed': Dictionary with 1 key
    'varied': DataFrame with 1 variable and 10 rows
'log': Dictionary with 5 keys
'measures': DataFrame with 1 variable and 10 rows
'variables': 
    'Agent': DataFrame with 1 variable and 10 rows
    'MyModel': DataFrame with 1 variable and 10 rows
}"""


class MyModel(ap.Model):
    def step(self):
        self.measure('x', self.p.x)
        self.add_agents()
        self.agents.record('id')
        self.record('id')
        self.stop()


def test_repr():
    param_ranges = {'x': (0., 1.), 'y': 1}
    sample = ap.sample(param_ranges, n=10)
    results = ap.Experiment(MyModel, sample, record=True).run()
    assert results.__repr__() == repr


class AgentType1(ap.Agent):
    def setup(self):
        self.x = 'x1'

    def action(self):
        self.record('x')


class AgentType2(AgentType1):
    def setup(self):
        self.x = 'x2'
        self.y = 'y2'

    def action(self):
        self.record(['x', 'y'])


class EnvType3(ap.Environment):
    def setup(self):
        self.x = 'x3'
        self.z = 'z4'

    def action(self):
        self.record(['x', 'z'])


class EnvType4(ap.Environment):
    def setup(self):
        self.z = 'z4'

    def action(self):
        self.record(['z'])


class ModelType0(ap.Model):

    def setup(self):
        self.E31 = self.add_env(env_class=EnvType3)
        self.E41 = self.add_env(env_class=EnvType4)
        self.E42 = self.add_env(env_class=EnvType4)
        self.envs.add_agents(agents=2, agent_class=AgentType1)
        self.E42.add_agents(agents=2, agent_class=AgentType2)

    def step(self):
        self.agents.action()
        self.envs.action()

    def end(self):
        self.measure('m_key', 'm_value')


def test_testing_model():

    parameters = {'steps': 2, 'px': (1, 2)}
    sample = ap.sample_discrete(parameters)
    settings = {'iterations': 2,
                'scenarios': ['test1', 'test2'],
                'record': True}

    pytest.model_instance = model = ModelType0(sample[0])
    pytest.model_results = model_results = model.run(display=False)

    exp = ap.Experiment(ModelType0, sample, **settings)
    pytest.exp_results = exp_results = exp.run(display=False)

    type_list = ['AgentType1', 'AgentType2', 'EnvType3', 'EnvType4']
    assert list(model_results.variables.keys()) == type_list
    assert list(exp_results.variables.keys()) == type_list


def arrange_things(results):

    return (results.arrange(variables='x'),
            results.arrange(variables=['x']),
            results.arrange(variables=['x', 'y']),
            results.arrange(variables='z'),
            results.arrange(parameters='px'),
            results.arrange(measures='m_key'),
            results.arrange(variables='all',
                            parameters='all',
                            measures='all'))


def test_datadict_arrange_for_single_run():

    results = pytest.model_results
    data = arrange_things(results)
    x_data, x_data2, xy_data, z_data, p_data, m_data, all_data = data

    assert x_data.equals(x_data2)
    assert list(x_data['x']) == ['x1'] * 4 + ['x2'] * 4 + ['x3'] * 2

    assert x_data.shape == (10, 4)
    assert xy_data.shape == (10, 5)
    assert z_data.shape == (6, 4)
    assert p_data.shape == (1, 2)
    assert m_data.shape == (1, 2)
    assert all_data.shape == (15, 10)


def test_datadict_arrange_for_multi_run():

    results = pytest.exp_results
    data = arrange_things(results)
    x_data, x_data2, xy_data, z_data, p_data, m_data, all_data = data

    assert x_data.equals(x_data2)
    assert x_data.shape == (80, 6)
    assert xy_data.shape == (80, 7)
    assert z_data.shape == (48, 6)
    assert p_data.shape == (4, 2)
    assert m_data.shape == (8, 3)
    assert all_data.shape == (120, 11)


def test_datadict_arrange_measures():

    results = pytest.exp_results
    mvp_data = results.arrange(measures='all', parameters='varied')
    mvp_data_2 = results.arrange_measures()
    assert mvp_data.equals(mvp_data_2)


def test_datadict_arrange_variables():

    results = pytest.exp_results
    mvp_data = results.arrange(variables='all', parameters='varied')
    mvp_data_2 = results.arrange_variables()
    assert mvp_data.equals(mvp_data_2)


def test_automatic_loading():

    if 'ap_output' in os.listdir():
        shutil.rmtree('ap_output')

    results = pytest.model_results
    results.log['test'] = False
    results.save(exp_name="a")
    results.save(exp_name="b", exp_id=1)
    results.log['test'] = True
    results.save(exp_name="b", exp_id=3)
    results.log['test'] = False
    results.save(exp_name="c")
    results.save(exp_name="b", exp_id=2)

    loaded = ap.load()
    shutil.rmtree('ap_output')

    # Latest experiment is chosen (b),
    # and then highest id is chosen (3)

    assert loaded.log['test'] is True


def test_saved_equals_loaded():

    results = pytest.exp_results
    results.save()
    loaded = ap.load('ModelType0')
    shutil.rmtree('ap_output')
    assert results == loaded
    # Test that equal doesn't hold if parts are changed
    assert results != 1
    loaded.measures = 1
    assert results != loaded
    results.measures = 1
    assert results == loaded
    loaded.log = 1
    assert results != loaded
    del loaded.log
    assert results != loaded


class WeirdObject:
    pass


def test_save_load():

    dd = ap.DataDict()
    dd['i1'] = 1
    dd['i2'] = np.int64(1)
    dd['f1'] = 1.
    dd['f2'] = np.float32(1.1)
    dd['s1'] = 'test'
    dd['s2'] = 'testtesttesttesttesttest'
    dd['l1'] = [1, 2, [3, 4]]
    dd['l2'] = np.array([1, 2, 3])
    dd['wo'] = WeirdObject()

    dd.save()
    dl = ap.load()
    with pytest.raises(FileNotFoundError):
        assert ap.load("Doesn't_exist")
    shutil.rmtree('ap_output')
    with pytest.raises(FileNotFoundError):
        assert ap.load("Doesn't_exist")

    assert dd.__repr__().count('\n') == 10
    assert dl.__repr__().count('\n') == 9
    assert len(dd) == 9
    assert len(dl) == 8
    assert dl.l1[2][1] == 4


def test_load_unreadable():
    """ Unreadable entries are loaded as None. """
    path = f'ap_output/fake_experiment_1/'
    os.makedirs(path)
    f = open(path + "unreadable_entry.xxx", "w+")
    f.close()
    dl = ap.load()
    shutil.rmtree('ap_output')
    assert dl.unreadable_entry is None




""" Agentpy Network Module """

import networkx as nx
from .objects import ApEnv, Agent
from .lists import AgentList
from .tools import make_list


class Network(ApEnv):
    """ Agent environment with a graph topology.
    Every node of the network represents an agent in the environment.
    To add new network environments to a model, use :func:`Model.add_network`.

    This class can be used as a parent class for custom network types.
    All agentpy model objects call the method :func:`setup` after creation,
    and can access class attributes like dictionary items.
    See :class:`Environment` for general properties of all environments.

    Arguments:
        model (Model): The model instance.
        graph (networkx.Graph, optional): The environments' graph.
            Agents of the same number as graph nodes must be passed.
            If none is passed, an empty graph is created.
        agents (AgentList, optional): Agents of the network (default None).
            If a graph is passed, agents are mapped to each node of the graph.
            Otherwise, new nodes will be created for each agent.
        **kwargs: Will be forwarded to :func:`Network.setup`.

    Attributes:
        graph (networkx.Graph): The environments' graph.
    """

    def __init__(self, model, graph=None, agents=None, **kwargs):

        super().__init__(model)

        if graph is None:
            self.graph = nx.Graph()
            if agents:
                self.add_agents(agents)
        elif isinstance(graph, nx.Graph):
            self.graph = graph
            # Map each agent to a node of the graph
            if agents is None or len(agents) != len(self.graph.nodes):
                la = len(agents) if agents else 0
                ln = len(self.graph.nodes)
                raise ValueError(
                    f"Number of agents ({la}) in 'agents' doesn't match "
                    f"number of nodes ({ln}) in graph.")
            super().add_agents(agents)  # Add agents without new nodes
            mapping = {i: agent for i, agent in enumerate(agents)}
            nx.relabel_nodes(self.graph, mapping=mapping, copy=False)
        else:
            raise TypeError("Argument 'graph' must be of type networkx.Graph")

        self._topology = 'network'
        self._set_var_ignore()
        self.setup(**kwargs)

    def add_agents(self, agents, agent_class=Agent, **kwargs):
        """ Adds agents to the network environment as new nodes.
        See :func:`Environment.add_agents` for standard arguments.
        """
        new_agents = super().add_agents(agents, agent_class, **kwargs)
        for agent in new_agents:
            self.graph.add_node(agent)  # Add agents to graph as new nodes
        return new_agents

    def remove_agents(self, agents):
        """ Removes agents from the environment. """
        for agent in make_list(agents):
            self.graph.remove_node(agent)
        super().remove_agents(agents)

    def neighbors(self, agent, **kwargs):
        """ Returns an :class:`AgentList` of agents
        that are connected to the passed agent. """
        return AgentList(
            [n for n in self.graph.neighbors(agent)], model=self.model)



"""
Agentpy Space Module
Content: Class for continuous spatial environments
"""

# There is much room for optimization in this module.
# Contributions are welcome! :)

# TODO Add option of space without shape (infinite)
# TODO Improve datatype consistency (list, np.array, etc.)
# TODO Method for distance between agents
# TODO Create method items() consistent with Grid

import itertools
import numpy as np
import random as rd
import collections.abc as abc
from scipy import spatial
from .objects import ApEnv, Agent
from .tools import make_list, make_matrix
from .lists import AgentList


class Space(ApEnv):
    """ Environment that contains agents with a continuous spatial topology.
    To add new space environments to a model, use :func:`Model.add_space`.
    For a discrete spatial topology, use :class:`Grid`.

    This class can be used as a parent class for custom space types.
    All agentpy model objects call the method :func:`setup` after creation,
    and can access class attributes like dictionary items.
    See :class:`Environment` for general properties of all environments.

    Arguments:
        model (Model): The model instance.
        shape (tuple of float): Size of the space.
            The length of the tuple defines the number of dimensions,
            and the values in the tuple define the length of each dimension.
        connect_borders (bool, optional):
            Whether to connect borders (default False).
            If True, space will be toroidal, meaning that agents who, for
            example, move over the right border, will appear on the left side.
        **kwargs: Will be forwarded to :func:`Space.setup`.

    Attributes:
        shape (tuple of float): Length of each spatial dimension.
        dim (int): Number of dimensions.
    """

    def __init__(self, model, shape, connect_borders=False, **kwargs):

        super().__init__(model)

        self._topology = 'space'
        self._connect_borders = connect_borders
        self._cKDTree = None
        self._sorted_agents = None
        self._sorted_agent_points = None
        self._agent_dict = {}  # agent instance : agent position (np.array)
        self._shape = tuple(shape)
        self._set_var_ignore()
        self.setup(**kwargs)

    @property
    def shape(self):
        return self._shape

    @property
    def dim(self):
        return len(self._shape)

    @property
    def KDTree(self):
        """ KDTree of agent positions for neighbor lookup,
        using :class:`scipy.spatial.KDTree`.
        Tree is recalculated if agent's have moved or changed.
        If there are no agents, returns None."""
        # Create new KDTree if necessary
        if self._cKDTree is None and len(self.agents) > 0:
            self._sorted_agents = []
            self._sorted_agent_points = []
            for a in self.agents:
                self._sorted_agents.append(a)
                self._sorted_agent_points.append(self._agent_dict[a])
            if self._connect_borders:
                self._cKDTree = spatial.cKDTree(self._sorted_agent_points,
                                                boxsize=self.shape)
            else:
                self._cKDTree = spatial.cKDTree(self._sorted_agent_points)
        return self._cKDTree  # Return existing or new KDTree

    def add_agents(self, agents=1, agent_class=Agent, positions=None,  # noqa
                   random=False, generator=None, **kwargs):
        """ Adds agents to the space environment, and returns new agents.
        See :func:`Environment.add_agents` for standard arguments.
        Additional arguments are listed below.

        Arguments:
            positions (array_like, optional):
                The positions of the added agents.
                Array must have the same length as number of agents
                to be added, and each entry must be an array with coordinates.
                If none is passed, agents will be placed in the
                bottom-left corner, i.e.: (0, 0, ...).
            random (bool, optional):
                If no positions are passed, agents will be placed in random
                locations instead of starting at the corner (default False).
            generator (numpy.random.Generator, optional):
                Random number generator.
                If none is passed, :obj:`Model.random` is used.
        """

        # Standard adding
        new_agents = super().add_agents(agents, agent_class, **kwargs)

        # Extra space features
        if not positions:
            n_agents = len(new_agents)
            if random:
                generator = generator if generator else self.model.random
                positions = [[generator.random() * d_max
                              for d_max in self.shape]
                             for _ in range(n_agents)]
            else:
                positions = [np.zeros(self.dim) for _ in range(n_agents)]

        for agent, pos in zip(new_agents, positions):
            self._agent_dict[agent] = np.array(pos)  # Add pos to agent_dict

        return new_agents

    def remove_agents(self, agents):
        """ Removes agents from the environment. """
        for agent in make_list(agents):
            del self._agent_dict[agent]
        super().remove_agents(agents)

    def position(self, agent):
        """ Returns :class:`numpy.array` with position of passed agent.

        Arguments:
            agent(int or Agent): Id or instance of the agent.
        """
        if isinstance(agent, int):
            agent = self.model.get_obj(agent)
        return self._agent_dict[agent]

    def positions(self, transpose=False):
        """ Returns list with positions of all agents.

        Arguments:
            transpose (bool, optional):
                If False (default), positions will be of style:
                [[agent1.x, agent1.y, ...], [agent2.x, ...], ...]
                If True, positions will be of style:
                [[agent1.x, agent2.x, ...], [agent1.y, ...], ...]
        """
        pos_per_agent = self._agent_dict.values()
        if not transpose:
            return pos_per_agent
        else:
            return map(list, zip(*pos_per_agent))

    def get_agents(self, center, radius):
        """ Returns an :class:`AgentList` with agents in selected search area,
        using :func:`scipy.cKDTree.query_ball_point`.

        Arguments:
            center (array_like): Coordinates of the center of the search area.
            radius (float): Radius around the center in which to search.
        """
        if self.KDTree:
            list_ids = self.KDTree.query_ball_point(center, radius)
            agents = [self._sorted_agents[list_id] for list_id in list_ids]
            return AgentList(agents, self.model)
        else:
            return AgentList(model=self.model)

    def neighbors(self, agent, distance):
        """ Returns an :class:`AgentList` with agent neighbors,
        using :func:`scipy.cKDTree.query_ball_point`.

        Arguments:
            agent(int or Agent): Id or instance of the agent.
            distance(float): Radius around the agent in which to
                search for neighbors.
        """

        list_ids = self.KDTree.query_ball_point(
            self._agent_dict[agent], distance)
        agents = [self._sorted_agents[list_id] for list_id in list_ids]
        agents.remove(agent)  # Remove original agent
        return AgentList(agents, self.model)

    def move_agent(self, agent, position):
        """ Moves agent to new position.

        Arguments:
            agent(int or Agent): Id or instance of the agent.
            position(array_like): New position of the agent.
        """

        if isinstance(agent, int):
            agent = self.model.get_obj(agent)

        for i in range(len(position)):

            # Border behavior
            if self._connect_borders:  # Connected
                while position[i] > self.shape[i]:
                    position[i] -= self.shape[i]
                while position[i] < 0:
                    position[i] += self.shape[i]
            else:  # Not connected - Stop at border
                if position[i] > self.shape[i]:
                    position[i] = self.shape[i]
                elif position[i] < 0:
                    position[i] = 0

        # Updating position
        self._agent_dict[agent] = np.array(position)
        self._cKDTree = None  # Reset KDTree



"""
Agentpy - Agent-based modeling in Python
Copyright (c) 2020 Joël Foramitti

Documentation: https://agentpy.readthedocs.io/
Examples: https://agentpy.readthedocs.io/en/latest/model_library.html
Source: https://github.com/JoelForamitti/agentpy
"""

__all__ = [
    '__version__',
    'Model', 'Environment', 'Network', 'Grid', 'Space',
    'Agent', 'AgentList', 'EnvList', 'ObjList', 'AttrList',
    'Experiment',
    'DataDict', 'load',
    'sample', 'sample_discrete', 'sample_saltelli',
    'sensitivity_sobol', 'gridplot', 'animate',
    'AttrDict'
]

# Version
try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python
    import importlib_metadata as metadata  # noqa
__version__ = metadata.version('agentpy')

# Objects
from .lists import AttrList, ObjList, AgentList, EnvList
from .objects import Agent, Environment
from .network import Network
from .grid import Grid
from .space import Space
from .model import Model
from .experiment import Experiment
from .output import DataDict, load
from .sampling import sample, sample_discrete, sample_saltelli
from .analysis import sensitivity_sobol, gridplot, animate
from .tools import AttrDict



import pytest
import agentpy as ap

repr = """Agent-based model {
'type': Model
'agents': AgentList [1 agent]
'envs': EnvList [1 environment]
'p': AttrDict {0 entries}
't': 0
'run_id': None
'scenario': None
'output': DataDict {1 entry}
}"""


def test_basics():
    model = ap.Model()
    model.add_env()
    model.env.add_agents()
    agent = model.agents[0]
    agent.x = 1
    agent['y'] = 2
    assert agent['x'] == 1
    assert agent.y == 2
    assert model.env.__repr__() == "Environment (Obj 1)"
    assert agent.__repr__() == "Agent (Obj 2)"
    assert model.env.topology is None
    assert model.type == 'Model'
    assert model.__repr__() == repr
    with pytest.raises(AttributeError):
        assert agent.z
    assert agent is model.get_obj(2)
    with pytest.raises(ValueError):
        assert model.get_obj(3)


def test_record():
    """ Record a dynamic variable """

    model = ap.Model()
    model.add_agents(3)
    model.var1 = 1
    model.var2 = 2
    model.record(['var1', 'var2'])
    model.record('var3', 3)

    assert len(list(model._log.keys())) == 3 + 1  # One for time
    assert model._log['var1'] == [1]
    assert model._log['var2'] == [2]
    assert model._log['var3'] == [3]


def test_record_all():
    """ Record all dynamic variables automatically """

    model = ap.Model()
    model.var1 = 1
    model.var2 = 2
    model.record(model.var_keys)

    assert len(list(model._log.keys())) == 3
    assert model._log['var1'] == [1]
    assert model._log['var2'] == [2]


"""
Agentpy Objects Module
Content: Classes for environments and agents
"""

from .lists import AgentList, EnvList
from .tools import AgentpyError, make_list


class ApObj:
    """ Agentpy base-class for objects of agent-based models."""

    def __init__(self, model):
        self._log = {}
        self._model = model
        self._envs = EnvList()
        self._var_ignore = []
        self._id = model._new_id()  # Assign id to new object
        self._model._obj_dict[self.id] = self  # Add object to object dict

    def __repr__(self):
        return f"{self.type} (Obj {self.id})"

    def __getattr__(self, key):
        raise AttributeError(f"{self} has no attribute '{key}'.")

    @property
    def type(self):
        """Class name of the object (str)."""
        return type(self).__name__

    @property
    def var_keys(self):
        """The object's variables (list of str)."""
        return [k for k in self.__dict__.keys()
                if k[0] != '_'
                and k not in self._var_ignore]

    @property
    def id(self):
        return self._id

    @property
    def p(self):
        return self._model._parameters

    @property
    def log(self):
        return self._log

    @property
    def model(self):
        return self._model

    @property
    def env(self):
        """ The objects first environment. """
        return self._envs[0] if self._envs else None

    @property
    def envs(self):
        return self._envs

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def _set_var_ignore(self):
        """Store current attributes to seperate them from custom variables"""
        self._var_ignore = [k for k in self.__dict__.keys() if k[0] != '_']

    def record(self, var_keys, value=None):
        """ Records an objects variables.

        Arguments:
            var_keys (str or list of str):
                Names of the variables to be recorded.
            value (optional): Value to be recorded.
                The same value will be used for all `var_keys`.
                If none is given, the values of object attributes
                with the same name as each var_key will be used.

        Examples:

            Record the existing attributes ``x`` and ``y`` of an object ``a``::

                a.record(['x', 'y'])

            Record a variable ``z`` with the value ``1`` for an object ``a``::

                a.record('z', 1)

            Record all variables of an object::

                a.record(a.var_keys)
        """

        for var_key in make_list(var_keys):

            # Create empty lists
            if 't' not in self.log:
                self.log['t'] = []
            if var_key not in self.log:
                self.log[var_key] = [None] * len(self.log['t'])

            if self.model.t not in self.log['t']:

                # Create empty slot for new documented time step
                for v in self.log.values():
                    v.append(None)

                # Store time step
                self.log['t'][-1] = self.model.t

            if value is None:
                v = getattr(self, var_key)
            else:
                v = value

            self.log[var_key][-1] = v

    def setup(self, **kwargs):
        """This empty method is called automatically at the objects' creation.
        Can be overwritten in custom sub-classes
        to define initial attributes and actions.

        Arguments:
            **kwargs: Keyword arguments that have been passed to
                :class:`Agent` or :func:`Model.add_agents`.
                If the original setup method is used,
                they will be set as attributes of the object.

        Examples:
            The following setup initializes an object with three variables::

                def setup(self, y):
                    self.x = 0  # Value defined locally
                    self.y = y  # Value defined in kwargs
                    self.z = self.p.z  # Value defined in parameters
        """

        for k, v in kwargs.items():
            setattr(self, k, v)


# Level 3 - Agent class

class Agent(ApObj):
    """ Individual agent of an agent-based model.

    This class can be used as a parent class for custom agent types.
    All agentpy model objects call the method :func:`setup()` after creation,
    and can access class attributes like dictionary items. To add new agents
    to a model, use :func:`Model.add_agents` or :func:`Environment.add_agents`.

    Arguments:
        model (Model): Instance of the current model.
        **kwargs: Will be forwarded to :func:`Agent.setup`.

    Attributes:
        model (Model): Model instance.
        p (AttrDict): Model parameters.
        envs (EnvList): Environments of the agent.
        log (dict): Recorded variables of the agent.
        id (int): Unique identifier of the agent instance.
    """

    def __init__(self, model, **kwargs):
        super().__init__(model)
        self._set_var_ignore()
        self.setup(**kwargs)

    def delete(self):
        """ Remove agent from all environments and the model. """
        for env in self.envs:
            env.remove_agents(self)
        self.model.remove_agents(self)

    def _find_env(self, env=None, topologies=None, new=False):
        """ Return obj of id or first object with topology. """
        # TODO Select based on method existance instead of topology
        if topologies:
            return self._find_env_top(env=env, topologies=topologies)
        if env is None:
            if self.envs:
                return self.envs[0]
            raise AgentpyError(f"{self} has no environments.")
        else:
            if isinstance(env, int):
                env = self.model.get_obj(env)
            if new or env in self.envs:
                return env
            raise AgentpyError(f"{self} is not part of environment {env}.")

    def _find_env_top(self, env=None, topologies=None):
        """ Return obj of id or first object with topology"""
        topologies = make_list(topologies)
        if env is None:
            for env in self.envs:
                if env.topology in topologies:
                    return env
            raise AgentpyError(
                f"Agent {self.id} has no environment "
                f"with topology '{topologies}'")
        else:
            if isinstance(env, int):
                env = self.model.get_obj(env)
            if hasattr(env, 'topology') and env.topology in topologies:
                return env
            raise AgentpyError(
                f"{env} does not have topology '{topologies}'")

    def position(self, env=None):
        """ Returns the agents' position from a grid.

        Arguments:
            env (int or Environment, optional):
                Instance or id of environment that should be used.
                Must have topology 'grid'.
                If none is given, the first environment of that topology
                in :attr:`Agent.envs` is used.
        """

        env = self._find_env(env, ['grid', 'space'])
        return env._agent_dict[self]

    def move_by(self, path, env=None):
        """ Changes the agents' location in the selected environment,
        relative to the current position.

        Arguments:
            path (list of int): Relative change of position.
            env (int or Environment, optional):
                Instance or id of environment that should be used.
                Must have topology 'grid'.
                If none is given, the first environment of that topology
                in :attr:`Agent.envs` is used.
        """
        # TODO Add border jumping feature (toroidal)
        env = self._find_env(env, ['grid', 'space'])
        old_pos = self.position(env)
        position = [p + c for p, c in zip(old_pos, path)]
        env.move_agent(self, position)

    def move_to(self, position, env=None):
        """ Changes the agents' location in the selected environment.

        Arguments:
            position (list of int): Position to move to.
            env(int or Environment, optional):
                Instance or id of environment that should be used.
                Must have topology 'grid'.
                If none is given, the first environment of that topology
                in :attr:`Agent.envs` is used.
        """

        env = self._find_env(env, ['grid', 'space'])
        env.move_agent(self, position)

    def neighbors(self, env=None, distance=1, **kwargs):
        """ Returns the agents' neighbor's from an environment,
        by calling the environments :func:`neighbors` function.

        Arguments:
            env(int or Environment, optional):
                Instance or id of environment that should be used.
                Must have topology 'grid' or 'network'.
                If none is given, the first environment of that topology
                in :attr:`Agent.envs` is used.
            distance(int, optional):
                Distance from agent in which to look for neighbors.
            **kwargs: Forwarded to the environments :func:`neighbors` function.

        Returns:
            AgentList: Neighbors of the agent.
        """
        env = self._find_env(env, ('grid', 'space', 'network'))
        return env.neighbors(self, distance=distance, **kwargs)

    def enter(self, env):
        """ Adds agent to passed environment.

        Arguments:
            env(int or Environment, optional):
                Instance or id of environment that should be used.
                If none is given, the first environment
                in :attr:`Agent.envs` is used.
        """
        env = self._find_env(env, new=True)
        env.add_agents(self)

    def exit(self, env=None):
        """ Removes agent from chosen environment.

        Arguments:
            env(int or Environment, optional):
                Instance or id of environment that should be used.
                If none is given, the first environment
                in :attr:`Agent.envs` is used.
        """
        env = self._find_env(env)
        env.remove_agents(self)


class ApEnv(ApObj):
    """ Agentpy base-class for environments. """

    def __init__(self, model):
        super().__init__(model)
        self._agents = AgentList(model=model)
        self._topology = None

    @property
    def topology(self):
        return self._topology

    @property
    def agents(self):
        return self._agents

    def remove_agents(self, agents):
        """ Removes agents from the environment. """
        is_env = True if self != self.model else False
        for agent in list(make_list(agents)):  # Soft copy
            self._agents.remove(agent)
            if is_env:
                agent.envs.remove(self)

    def add_agents(self, agents=1, agent_class=Agent, **kwargs):
        """ Adds agents to the environment.

        Arguments:
            agents(int or AgentList, optional): Either number of new agents
                to be created or list of existing agents (default 1).
            agent_class(type, optional): Type of new agents to be created
                if int is passed for agents (default :class:`Agent`).
            **kwargs: Forwarded to :func:`Agent.setup` if new agents are
                created (i.e. if an integer number is passed to `agents`).

        Returns:
            AgentList: List of the new agents.
        """

        # Check if object is environment or model
        is_env = True if self != self.model else False

        # Case 1 - Create new agents
        if isinstance(agents, int):
            agents = AgentList([agent_class(self.model, **kwargs)
                                for _ in range(agents)], model=self.model)
            if is_env:  # Add agents to master list
                self.model._agents.extend(agents)

        # Case 2 - Add existing agents
        else:
            if not isinstance(agents, AgentList):
                agents = AgentList(make_list(agents), model=self.model)

        # Add environment to agents
        if is_env:
            for agent in agents:
                agent.envs.append(self)

        # Add agents to environment
        self._agents.extend(agents)

        return agents


class Environment(ApEnv):
    """ Standard environment for agents (no topology).

    This class can be used as a parent class for custom environment types.
    All agentpy model objects call the method :func:`setup()` after creation,
    and can access class attributes like dictionary items. To add new
    environments to a model, use :func:`Model.add_env`.

    Arguments:
        model (Model): The model instance.
        **kwargs: Will be forwarded to :func:`Environment.setup`.

    Attributes:
        model (Model): Model instance.
        p (AttrDict): Model parameters.
        id (int): Unique identifier of the environment instance.
        topology (str): Topology of the environment.
        log (dict): The environments' recorded variables.
    """

    def __init__(self, model, **kwargs):
        super().__init__(model)
        self._set_var_ignore()
        self.setup(**kwargs)



"""
Agentpy Sampling Module
Content: Sampling functions
"""

import itertools
import numpy as np

from SALib.sample import saltelli
from .tools import param_tuples_to_salib

def sample(parameter_ranges, n, digits=None):
    """ Creates a sample of different parameter combinations
    by seperating each range into 'n' values, using :func:`numpy.linspace`.

    Arguments:
        parameter_ranges(dict): Dictionary of parameters.
            Only values that are given as a tuple will be varied.
            Tuple must be of the following style: (min_value, max_value).
            If both values are of type int,
            the output will be rounded and converted to int.
        n(int): Number of values to sample per varied parameter.
        digits(int, optional):
            Number of digits to round the output values to (default None).

    Returns:
        list of dict: List of parameter dictionaries
    """

    parameter_ranges = dict(parameter_ranges)
    for k, v in parameter_ranges.items():
        if isinstance(v, tuple):
            p_range = np.linspace(v[0], v[1], n)
            if all([isinstance(pv, int) for pv in v[:2]]):
                p_range = [int(round(pv)) for pv in p_range]
            elif digits is not None:
                p_range = [round(pv, digits) for pv in p_range]
            parameter_ranges[k] = p_range
        else:
            parameter_ranges[k] = [v]

    param_combinations = list(itertools.product(*parameter_ranges.values()))
    param_sample = [{k: v for k, v in zip(parameter_ranges.keys(), parameters)}
                    for parameters in param_combinations]

    return param_sample


def sample_discrete(parameter_ranges):
    """ Creates a sample of different parameter combinations from all possible
    combinations within the passed parameter ranges.

    Arguments:
        parameter_ranges (dict): Dictionary of parameters.
            Only values that are given as a tuple will be varied.
            Tuples must be of the following style:
            (value1, value2, value3, ...).

    Returns:
        list of dict: List of parameter dictionaries
    """

    def make_tuple(v):
        if isinstance(v, tuple):
            return v
        else:
            return (v,)

    param_ranges_values = [make_tuple(v) for k, v in parameter_ranges.items()]
    param_combinations = list(itertools.product(*param_ranges_values))
    param_sample = [{k: v for k, v in zip(parameter_ranges.keys(), parameters)}
                    for parameters in param_combinations]

    return param_sample


def _is_int(param_tuples):
    is_int = {}
    for key, p_values in param_tuples.items():
        is_int[key] = all([isinstance(v, int) for v in p_values[:2]])
    return is_int


def sample_saltelli(parameter_ranges, n, calc_second_order=False, digits=None):
    """ Creates a sample of different parameter combinations,
    using :func:`SALib.sample.saltelli.sample`. This sample can later be used
    to calculate Sobol sensitivity indices with :func:`sensitivity_sobol`.

    Arguments:
        parameter_ranges (dict): Dictionary of parameters.
            Only values that are given as a tuple will be varied.
            Tuple must be of the following style: (min_value, max_value).
            If both values are of type int,
            the output will be rounded and converted to int.
        n (int): The number of samples to generate,
            see :func:`SALib.sample.saltelli.sample`.
        calc_second_order (bool, optional):
            Create sample that can be used by :func:`sensitivity_sobol`
            to calculate second-order sensitivities (default False).
        digits (int, optional):
            Number of digits to round the output values to (default None).

    Returns:
        list of dict: List of parameter dictionaries
    """

    # STEP 0 - Find variable parameters and check type
    param_ranges_tuples = {k: v for k, v in parameter_ranges.items()
                           if isinstance(v, tuple)}

    # STEP 1 - Convert param_ranges to SALib Format
    param_ranges_salib = param_tuples_to_salib(param_ranges_tuples)

    # STEP 2 - Create SALib Sample
    salib_sample = saltelli.sample(param_ranges_salib, n, calc_second_order)

    # STEP 3 - Convert back to Agentpy Parameter Dict List and adjust values
    ap_sample = []
    is_int = _is_int(param_ranges_tuples)

    for param_instance in salib_sample:

        parameters = {}
        parameters.update(parameter_ranges)

        for i, key in enumerate(param_ranges_tuples.keys()):
            p = param_instance[i]
            if is_int[key]:  # Convert to integer
                p = int(round(p))
            elif digits:  # Round to digits
                p = round(p, digits)
            parameters[key] = p

        ap_sample.append(parameters)

    return ap_sample



import pytest
import agentpy as ap
import numpy as np


def test_repr():
    model = ap.Model()
    model.add_agents()
    model.add_env()
    assert model.agents.__repr__() == "AgentList [1 agent]"
    assert model.envs.__repr__() == "EnvList [1 environment]"
    assert model.objects.__repr__() == "ObjList [2 objects]"
    l1 = model.agents.id
    l2 = l1 + 1
    assert l1.__repr__() == "AttrList of attribute 'id': [1]"
    assert l2.__repr__() == "AttrList: [2]"


def test_attr_calls():
    model = ap.Model()
    model.add_agents(2)
    model.agents.x = 1
    model.agents.f = lambda: 2
    assert list(model.agents.x) == [1, 1]
    assert list(model.agents.f()) == [2, 2]
    with pytest.raises(AttributeError):
        assert model.agents.y
    with pytest.raises(TypeError):
        assert model.agents.x()  # noqa


def test_select():
    """ Select subsets with boolean operators. """
    model = ap.Model()
    model.add_agents(3)
    selection1 = model.agents.id == 2
    selection2 = model.agents.id != 2
    selection3 = model.agents.id < 2
    selection4 = model.agents.id > 2
    selection5 = model.agents.id <= 2
    selection6 = model.agents.id >= 2
    assert selection1 == [False, True, False]
    assert selection2 == [True, False, True]
    assert selection3 == [True, False, False]
    assert selection4 == [False, False, True]
    assert selection5 == [True, True, False]
    assert selection6 == [False, True, True]
    assert model.agents(selection1) == model.agents.select(selection1)
    assert list(model.agents(selection1).id) == [2]


def test_random():
    """ Test random shuffle and selection. """
    model = ap.Model()
    model.add_agents(2)
    assert len(model.agents) == len(model.agents.shuffle())
    assert len(model.agents.random()) == 1

    # Custom generator with seperate seed
    model = ap.Model()
    model.add_agents(5)
    generator = np.random.default_rng(1)
    assert len(model.agents.random(generator=generator)) == 1
    assert model.agents.random(generator=generator).id[0] == 3
    assert list(model.agents.shuffle(generator=generator).id) == [5, 1, 3, 2, 4]

    # Test with single agent
    model = ap.Model()
    agents = model.add_agents(1)
    assert model.agents.random()[0] is agents[0]
    assert model.agents.shuffle()[0] is agents [0]

    # Agentlist with no model defined directly
    model = ap.Model()
    agents = model.add_agents(3)
    agents = ap.AgentList(agents)
    model.run(steps=0, seed=1, display=False)
    assert agents.random()[0].id == 2

    # Agentlist with no model defined
    # (no seed control without model, test can only check if no errors)
    agents1 = ap.AgentList([1, 2, 3])
    agents1.random()


def test_sort():
    """ Test sorting method. """
    model = ap.Model()
    model.add_agents(2)
    model.agents[0].x = 1
    model.agents[1].x = 0
    model.agents.sort('x')
    assert list(model.agents.x) == [0, 1]
    assert list(model.agents.id) == [2, 1]


def test_arithmetics():
    """ Test arithmetic operators """

    model = ap.Model()
    model.add_agents(3)
    agents = model.agents

    agents.x = 1
    assert agents.x.attr == "x"
    assert list(agents.x) == [1, 1, 1]

    agents.y = ap.AttrList([1, 2, 3])
    assert list(agents.y) == [1, 2, 3]

    agents.x = agents.x + agents.y
    assert list(agents.x) == [2, 3, 4]

    agents.x = agents.x - ap.AttrList([1, 1, 1])
    assert list(agents.x) == [1, 2, 3]

    agents.x += 1
    assert list(agents.x) == [2, 3, 4]

    agents.x -= 1
    assert list(agents.x) == [1, 2, 3]

    agents.x *= 2
    assert list(agents.x) == [2, 4, 6]

    agents.x = agents.x * agents.x
    assert list(agents.x) == [4, 16, 36]

    agents.x = agents.x / agents.x
    assert list(agents.x)[0] == pytest.approx(1.)

    agents.x /= 2
    assert list(agents.x)[0] == pytest.approx(0.5)



import pytest
import agentpy as ap
import numpy as np


def test_general():

    model = ap.Model()
    grid = model.add_grid((2, 2))
    assert grid is model.env
    assert grid.shape == (2, 2)
    assert grid.dim == 2

    # 5th agent must be at first position again
    model.env.add_agents(5)
    assert len(model.env.grid[0][0]) == 2
    assert len(model.env.grid[0][1]) == 1

    # Test only for errors
    model.env.add_agents(5, random=True)


def test_remove():
    model = ap.Model()
    grid = model.add_grid((2, 2))
    grid.add_agents(2)
    agent = model.agents[0]
    grid.remove_agents(agent)
    assert grid.attribute('id') == [[np.nan, 3], [np.nan, np.nan]]
    assert len(grid._agent_dict) == 1
    assert len(grid.agents) == 1


def test_positions():
    model = ap.Model()
    grid = model.add_grid((2, 2))
    grid.add_agents(5)

    # Position reference
    a = model.agents[0]
    assert a.position() == (0, 0)
    assert a.position() is grid.position(a)  # By instance
    assert a.position() is grid.position(a.id)  # By id

    # Positions
    assert list(grid.positions()) == [(0, 0), (0, 1), (1, 0), (1, 1)]
    assert list(grid.positions([(0, 0), (0, 1)])) == [(0, 0), (0, 1)]

    # Get agents
    assert len(grid.get_agents()) == 5
    assert len(grid.get_agents([0, 0])) == 2  # Single pos
    assert len(grid.get_agents([(0, 0), (0, 1)])) == 3  # Area

    # Get items
    assert [x for x in grid.items()][0][0] == (0, 0)
    assert len([x for x in grid.items()][0][1]) == 2

    # Wrong area
    with pytest.raises(ValueError):
        assert grid.get_agents(1)


def test_attribute():
    model = ap.Model()
    model.add_grid((2, 2))
    model.env.add_agents(5)
    assert model.env.attribute('id') == [[8, 3], [4, 5]]
    assert model.env.attribute('id', sum_values=False) == [[[2, 6], [3]],
                                                           [[4], [5]]]


def test_apply():

    model = ap.Model()
    model.add_grid((3, 3))
    model.env.add_agents(1, positions=[(1, 1)])
    # Apply function len to each position, must be 1 where agent is
    assert model.env.apply(len) == [[0, 0, 0], [0, 1, 0], [0, 0, 0]]


def test_movement():

    model = ap.Model()
    grid = model.add_grid((2, 2))
    grid.add_agents(2)
    agent = model.get_obj(2)

    assert grid.attribute('id') == [[2, 3], [np.nan, np.nan]]
    agent.move_to((1, 0))  # Move in absolute terms
    assert grid.attribute('id') == [[np.nan, 3], [2, np.nan]]
    agent.move_by((-1, 0), env=grid.id)  # Move in relative terms
    assert grid.attribute('id') == [[2, 3], [np.nan, np.nan]]
    grid.move_agent(2, (1, 0))  # Move through grid call by id
    assert grid.attribute('id') == [[np.nan, 3], [2, np.nan]]

    assert agent.neighbors() == [model.get_obj(3)]


def test_neighbors():

    model = ap.Model()
    grid = model.add_grid((5, 5))
    grid.add_agents(25)
    grid.attribute('id')
    a = model.agents[12]

    assert a.neighbors() == grid.neighbors(a.id)
    assert len(a.neighbors(distance=1)) == 8
    assert len(a.neighbors(distance=2)) == 24
    assert len(a.neighbors(diagonal=False)) == 4



import pytest
import numpy as np
import agentpy as ap

from agentpy.tools import AgentpyError


def test_run():
    """ Test time limits. """

    # Parameter step limit
    model = ap.Model()
    assert model.t == 0
    model.p.steps = 0
    model.run()
    assert model.t == 0
    model.p.steps = 1
    model.run()
    assert model.t == 1

    # Maximum time limit
    del model.p.steps
    model.t = 999
    model.run()
    assert model.t == 1000


def test_run_seed():
    """ Test random seed setting. """
    n = np.random.default_rng(1).integers(10)
    model = ap.Model({'seed': 1})
    model.run(steps=0, display=False)
    assert model.random.integers(10) == n
    model = ap.Model()
    model.run(seed=1, steps=0, display=False)
    assert model.random.integers(10) == n


def test_stop():
    """ Test method Model.stop() """

    class Model(ap.Model):
        def step(self):
            if self.t == 2:
                self.stop()

    model = Model()
    model.run()

    assert model.t == 2


def test_add_agents():
    """ Add new agents to model """

    model = ap.Model()
    model.add_agents(3)

    assert len(model.agents) == 3  # New agents
    assert list(model.agents.id) == [1, 2, 3]

    model.add_agents(model.agents)  # Existing agents
    assert list(model.agents.id) == [1, 2, 3] * 2


def test_objects_property():

    model = ap.Model()
    model.add_agents(3)
    model.add_env()

    assert len(model.objects) == 4
    assert model.agents[0] in model.objects
    assert model.envs[0] in model.objects


def test_setup():
    """ Test setup() for all object types """

    class MySetup:
        def setup(self, a):
            self.a = a + 1

    class MyAgentType(MySetup, ap.Agent):
        pass

    class MyEnvType(MySetup, ap.Environment):
        pass

    class MyNwType(MySetup, ap.Network):
        pass

    class MyGridType(MySetup, ap.Grid):
        pass

    model = ap.Model()
    model.add_agents(1, b=1)
    model.add_agents(1, MyAgentType, a=1)
    model.E1 = model.add_env(MyEnvType, a=2)
    model.G1 = model.add_env(MyGridType, shape=(1, 1), a=3)
    model.N1 = model.add_env(MyNwType, a=4)

    # Standard setup implements keywords as attributes
    # Custom setup uses only keyword a and adds 1

    with pytest.raises(TypeError):
        assert model.add_agents(1, MyAgentType, b=1)

    assert model.agents[0].b == 1
    assert model.agents[1].a == 2
    assert model.E1.a == 3
    assert model.G1.a == 4
    assert model.N1.a == 5


def test_delete():
    """ Remove agent from model """

    model = ap.Model()
    model.add_agents(3)
    model.add_env().add_agents(model.agents)
    model.agents[1].delete()

    assert len(model.agents) == 2
    assert list(model.agents.id) == [1, 3]
    assert list(model.env.agents.id) == [1, 3]


def test_create_output():
    """ Should put variables directly into output if there are only model
    variables, or make a subdict if there are also other variables. """

    model = ap.Model()
    model.record('x', 0)
    model.run(1)
    assert list(model.output.variables.keys()) == ['x']

    model = ap.Model(run_id=1, scenario='test')
    model.add_agents()
    model.agents.record('x', 0)
    model.record('x', 0)
    model.run(1)
    assert list(model.output.variables.keys()) == ['Agent', 'Model']

    # Run id and scenario should be added to output
    assert model.output.variables.Model.reset_index()['run_id'][0] == 1
    assert model.output.variables.Model.reset_index()['scenario'][0] == 'test'



"""
Agentpy Output Module
Content: DataDict class for output data
"""

import pandas as pd
import os
from os import listdir, makedirs
from os.path import getmtime, join

from .tools import AttrDict, make_list, AgentpyError
import json
import numpy as np


class NpEncoder(json.JSONEncoder):
    """ Adds support for numpy number formats to json. """
    # By Jie Yang https://stackoverflow.com/a/57915246
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)


def _last_exp_id(name, path):
    """ Identifies existing experiment data and return highest id. """

    exp_id = 0
    output_dirs = listdir(path)
    exp_dirs = [s for s in output_dirs if name in s]
    if exp_dirs:
        ids = [int(s.split('_')[-1]) for s in exp_dirs]
        exp_id = max(ids)
    return exp_id


class DataDict(AttrDict):
    """ Dictionary for recorded simulation data.

    Generated by :class:`Model`, :class:`Experiment`, or :func:`load`.
    Dictionary items can be defined and accessed like attributes.
    Attributes can differ from the standard ones listed below.

    Attributes:
        log (dict): Meta-data of the simulation
            (e.g. name, time-stamps, settings, etc.).
        parameters (dict, pandas.DataFrame, or DataDict):
            Parameters that have been used for the simulation.
        variables (pandas.DataFrame or DataDict)):
            Dynamic variables, seperated per object type,
            which can be recorded once per time-step with :func:`record`.
        measures (pandas.DataFrame): Evaluation measures,
            which can be recorded once per run with :func:`measure`.
    """

    def __repr__(self, indent=False):
        rep = ""
        if not indent:
            rep += "DataDict {"
        i = '    ' if indent else ''
        for k, v in self.items():
            rep += f"\n{i}'{k}': "
            if isinstance(v, (int, float, np.integer, np.floating)):
                rep += f"{v} {type(v)}"
            elif isinstance(v, str):
                x0 = f"(length {len(v)})"
                x = f"...' {x0}" if len(v) > 20 else "'"
                rep += f"'{v[:30]}{x} {type(v)}"
            elif isinstance(v, pd.DataFrame):
                lv = len(list(v.columns))
                rv = len(list(v.index))
                rep += f"DataFrame with {lv} " \
                       f"variable{'s' if lv != 1 else ''} " \
                       f"and {rv} row{'s' if rv != 1 else ''}"
            elif isinstance(v, DataDict):
                rep += f"{v.__repr__(indent=True)}"
            elif isinstance(v, dict):
                lv = len(list(v.keys()))
                rep += f"Dictionary with {lv} key{'s' if lv != 1 else ''}"
            elif isinstance(v, list):
                lv = len(v)
                rep += f"List with {lv} entr{'ies' if lv != 1 else 'y'}"
            else:
                rep += f"Object of type {type(v)}"
        if not indent:
            rep += "\n}"
        return rep

    def _short_repr(self):
        len_ = len(self.keys())
        return f"DataDict {{{len_} entr{'y' if len_ == 1 else 'ies'}}}"

    def __eq__(self, other):
        """ Check equivalence of two DataDicts."""
        if not isinstance(other, DataDict):
            return False
        for key, item in self.items():
            if key not in other:
                return False
            if isinstance(item, pd.DataFrame):
                if not self[key].equals(other[key]):
                    return False
            elif not self[key] == other[key]:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def _combine_vars(self, obj_types='all', var_keys='all'):
        """ Returns pandas dataframe with combined variables """

        # Retrieve variables
        if 'variables' in self:
            vs = self['variables']
        else:
            return None
        if isinstance(vs, pd.DataFrame):
            return vs  # Return df if vs is already a df
        elif isinstance(vs, DataDict) and len(vs.keys()) == 1:
            return list(vs.values())[0]  # Return df if vs has only one entry
        elif isinstance(vs, (dict,DataDict)):
            df_dict = dict(vs)  # Convert to dict if vs is DataDict
        else:
            raise TypeError("DataDict.variables must be of type dict,"
                             "agentpy.DataDict, or pandas.DataFrame.")

        # Remove dataframes that don't include any of the selected var_keys
        if var_keys != 'all':
            df_dict = {k: v for k, v in df_dict.items()
                       if any(x in v.columns for x in make_list(var_keys))}

        # Select object types
        if obj_types != 'all':
            df_dict = {k: v for k, v in df_dict.items()
                       if k in make_list(obj_types)}

        # Add 'obj_id' before 't' for model df
        model_type = self.log['model_type']
        if model_type in list(df_dict.keys()):
            df = df_dict[model_type]
            df['obj_id'] = 0
            indexes = list(df.index.names)
            indexes.insert(-1, 'obj_id')
            df = df.reset_index()
            df = df.set_index(indexes)
            df_dict[model_type] = df

        # Return none if empty
        if df_dict == {}:
            return None

        # Create dataframe
        df = pd.concat(df_dict)  # Dict keys (obj_type) will be added to index
        df.index = df.index.set_names('obj_type', level=0)  # Rename new index

        # Select var_keys
        if var_keys != 'all':
            # make_list prevents conversion to pd.Series for single value
            df = df[make_list(var_keys)]

        return df

    def _combine_pars(self, varied=True, fixed=True):
        """ Returns pandas dataframe with parameters and run_id """
        # Case 0: Cancel if there are no parameters
        if 'parameters' not in self:
            return None
        # Case 1: There is a subdict with fixed & combined
        dfp = None
        if isinstance(self.parameters, DataDict):
            dfp = pd.DataFrame()
            if varied:
                dfp = self.parameters.varied.copy()
            if fixed:
                for k, v in self.parameters.fixed.items():
                    dfp[k] = v
        # Case 2: There is a dict with fixed parameters
        elif isinstance(self.parameters, dict):
            if fixed:
                dfp = pd.DataFrame({k: [v] for k, v in self.parameters.items()})
        # Case 3: There is a dataframe with varied parameters
        elif isinstance(self.parameters, pd.DataFrame):
            if varied:
                dfp = self.parameters.copy()
        # Case 4: No parameters have been selected
        else:
            raise TypeError("DataDict.parameters must be of type"
                            "DataDict, dict, or pandas.DataFrame.")
        # Case 5: Cancel if no parameters have been selected
        if dfp is None or dfp.shape == (0, 0):
            return None
        # Case 1-3: Multiply for iterations, set new index, and return
        if 'iterations' in self.log and self.log['iterations'] > 1:
            dfp = pd.concat([dfp] * self.log['iterations'])
        dfp = dfp.reset_index(drop=True)
        dfp.index.name = 'run_id'
        return dfp

    def arrange_measures(self, variables=None, measures='all',
                         parameters='varied', obj_types='all',
                         scenarios='all', index=False):
        """ Returns a dataframe with measures and varied parameters.
        See :func:`DataDict.arrange` for further information."""
        return self.arrange(variables=variables, measures=measures,
                            parameters=parameters, obj_types=obj_types,
                            scenarios=scenarios, index=index)

    def arrange_variables(self, variables='all', measures=None,
                          parameters='varied', obj_types='all',
                          scenarios='all', index=False):
        """ Returns a dataframe with variables and varied parameters.
        See :func:`DataDict.arrange` for further information."""
        return self.arrange(variables=variables, measures=measures,
                            parameters=parameters, obj_types=obj_types,
                            scenarios=scenarios, index=index)

    def arrange(self, variables=None, measures=None, parameters=None,
                obj_types='all', scenarios='all', index=False):
        """ Combines and/or filters data based on passed arguments.

        Arguments:
            variables (str or list of str, optional):
                Variables to include in the new dataframe (default None).
                If 'all', all are selected.
            measures (str or list of str, optional):
                Measures to include in the new dataframe (default None).
                If 'all', all are selected.
            parameters (str or list of str, optional):
                Parameters to include in the new dataframe (default None).
                If 'fixed', all fixed parameters are selected.
                If 'varied', all varied parameters are selected.
                If 'all', all are selected.
            obj_types (str or list of str, optional):
                Agent and/or environment types to include in the new dataframe.
                Note that the selected object types will only be included
                if at least one of their variables is declared in 'variables'.
                If 'all', all are selected (default).
            scenarios (str or list of str, optional):
                Scenarios to include in the new dataframe.
                If 'all', all are selected (default).
            index (bool, optional):
                Whether to keep original multi-index structure (default False).

        Returns:
            pandas.DataFrame: The arranged dataframe
        """

        dfv = dfm = dfp = df = None

        # Step 1: Variables
        if variables is not None:
            dfv = self._combine_vars(obj_types, variables)

        # Step 2: Measures
        if measures is not None:
            dfm = self.measures
            if measures != 'all':  # Select measure keys
                # make_list prevents conversion to pd.Series for single value
                dfm = dfm[make_list(measures)]

        # Step 3: Parameters
        if parameters is not None:
            varied = False if parameters == 'static' else True
            static = False if parameters == 'varied' else True
            dfp = self._combine_pars(varied, static)
            if parameters not in ['all', 'varied', 'static']:
                # Select parameter keys
                # make_list prevents conversion to pd.Series for single value
                dfp = dfp[make_list(parameters)]

        # Step 4: Combine dataframes
        if dfv is not None and dfm is not None:
            # Combine variables & measures
            index_keys = dfv.index.names
            dfm = dfm.reset_index()
            dfv = dfv.reset_index()
            df = pd.concat([dfm, dfv])
            df = df.set_index(index_keys)
        elif dfv is not None:
            df = dfv
        elif dfm is not None:
            df = dfm
        if dfp is not None:
            if df is None:
                df = dfp
            else:  # Combine df with parameters
                if len(dfp) > 1:  # If multi run, add parameters by run_id
                    if df is not None and isinstance(df.index, pd.MultiIndex):
                        dfp = dfp.reindex(df.index, level='run_id')
                    df = pd.concat([df, dfp], axis=1)
                else:  # Elif single run, add parameters as columns
                    for k, v in dfp.items():
                        # dfp is a dataframe, items returns columns, Series
                        df[k] = v[0]
        if df is None:
            return None

        # Step 5: Select scenarios
        if scenarios != 'all' and 'scenario' in df.index.names:
            scenarios = make_list(scenarios)  # noqa
            df = df.query("scenario in @scenarios")

        # Step 6: Reset index
        if not index:
            df = df.reset_index()

        return df

    def save(self, exp_name=None, exp_id=None, path='ap_output', display=True):

        """ Writes data to directory `{path}/{exp_name}_{exp_id}/`.
        Works only for entries that are of type :class:`DataDict`,
        :class:`pandas.DataFrame`, or serializable with JSON
        (int, float, str, dict, list). Numpy objects will be converted
        to standard objects, if possible.

        Arguments:
            exp_name (str, optional): Name of the experiment to be saved.
                If none is passed, `self.log['name']` is used.
            exp_id (int, optional): Number of the experiment.
                If none is passed, a new id is generated.
            path (str, optional): Target directory (default 'ap_output').
            display (bool, optional): Display saving progress (default True).

        """

        # Create output directory if it doesn't exist
        if path not in listdir():
            makedirs(path)

        # Set exp_name
        if exp_name is None:
            if 'log' in self and 'name' in self.log:
                exp_name = self.log['name']
            else:
                exp_name = 'Unnamed'

        exp_name = exp_name.replace(" ", "_")

        # Set exp_id
        if exp_id is None:
            exp_id = _last_exp_id(exp_name, path) + 1

        # Create new directory for output
        path = f'{path}/{exp_name}_{exp_id}'
        makedirs(path)

        # Save experiment data
        for key, output in self.items():

            if isinstance(output, pd.DataFrame):
                output.to_csv(f'{path}/{key}.csv')

            elif isinstance(output, DataDict):
                for k, o in output.items():

                    if isinstance(o, pd.DataFrame):
                        o.to_csv(f'{path}/{key}_{k}.csv')
                    elif isinstance(o, dict):
                        with open(f'{path}/{key}_{k}.json', 'w') as fp:
                            json.dump(o, fp, cls=NpEncoder)

            else:  # Use JSON for other object types
                try:
                    with open(f'{path}/{key}.json', 'w') as fp:
                        json.dump(output, fp, cls=NpEncoder)
                except TypeError as e:
                    print(f"Warning: Object '{key}' could not be saved. "
                          f"(Reason: {e})")
                    os.remove(f'{path}/{key}.json')

            # TODO Support grids & graphs
            # elif t == nx.Graph:
            #    nx.write_graphml(output, f'{path}/{key}.graphml')

        if display:
            print(f"Data saved to {path}")

    def _load(self, exp_name=None, exp_id=None,
              path='ap_output', display=True):

        def load_file(path, file, display):
            if display:
                print(f'Loading {file} - ', end='')
            i_cols = ['sample_id', 'run_id', 'scenario',
                      'env_key', 'agent_id', 'obj_id', 't']
            ext = file.split(".")[-1]
            path = path + file
            try:
                if ext == 'csv':
                    obj = pd.read_csv(path) # Convert .csv into DataFrane
                    index = [i for i in i_cols if i in obj.columns]
                    if index:  # Set potential index columns
                        obj = obj.set_index(index)
                elif ext == 'json':
                    # Convert .json with json decoder
                    with open(path, 'r') as fp:
                        obj = json.load(fp)
                    # Convert dict to AttrDict
                    if isinstance(obj, dict):
                        obj = AttrDict(obj)
                # TODO Support grids & graphs
                # elif ext == 'graphml':
                #    self[key] = nx.read_graphml(path)
                else:
                    raise ValueError(f"File type '{ext}' not supported")
                if display:
                    print('Successful')
                return obj
            except Exception as e:
                print(f'Error: {e}')

        # Prepare for loading
        if exp_name is None:
            # Choose latest modified experiment
            exp_names = listdir(path)
            paths = [join(path, d) for d in exp_names]
            latest_exp = exp_names[paths.index(max(paths, key=getmtime))]
            exp_name = latest_exp.rsplit('_', 1)[0]

        exp_name = exp_name.replace(" ", "_")
        if not exp_id:
            exp_id = _last_exp_id(exp_name, path)
            if exp_id == 0:
                raise FileNotFoundError(f"No experiment found with "
                                        f"name '{exp_name}' in path '{path}'")
        path = f'{path}/{exp_name}_{exp_id}/'
        if display:
            print(f'Loading from directory {path}')

        # Loading data
        for file in listdir(path):
            if 'variables_' in file:
                if 'variables' not in self:
                    self['variables'] = DataDict()
                ext = file.split(".")[-1]
                key = file[:-(len(ext) + 1)].replace('variables_', '')
                self['variables'][key] = load_file(path, file, display)
            elif 'parameters_' in file:
                ext = file.split(".")[-1]
                key = file[:-(len(ext) + 1)].replace('parameters_', '')
                if 'parameters' not in self:
                    self['parameters'] = DataDict()
                self['parameters'][key] = load_file(path, file, display)
            else:
                ext = file.split(".")[-1]
                key = file[:-(len(ext) + 1)]
                self[key] = load_file(path, file, display)
        return self


def load(exp_name=None, exp_id=None, path='ap_output', display=True):
    """ Reads output data from directory `{path}/{exp_name}_{exp_id}/`.

        Arguments:
            exp_name (str, optional): Experiment name.
                If none is passed, the most recent experiment is chosen.
            exp_id (int, optional): Id number of the experiment.
                If none is passed, the highest available id used.
            path (str, optional): Target directory (default 'ap_output').
            display (bool, optional): Display loading progress (default True).

        Returns:
            DataDict: The loaded data from the chosen experiment.
    """
    return DataDict()._load(exp_name, exp_id, path, display)



import pytest
import agentpy as ap
from SALib.sample import saltelli


out = [{'a': 1, 'b': 3.0, 'c': 5},
       {'a': 1, 'b': 3.5, 'c': 5},
       {'a': 2, 'b': 3.0, 'c': 5},
       {'a': 2, 'b': 3.5, 'c': 5}]


def test_sample():
    parameters = {'a': (1, 2), 'b': (3., 3.5), 'c': 5}
    sample = ap.sample(parameters, n=2)
    assert sample == out


def test_rounded_sample():
    parameters = {'a': (1, 2), 'b': (3., 3.5), 'c': 5}
    sample = ap.sample(parameters, n=2, digits=0)
    out[1]['b'] = out[3]['b'] = 4
    assert sample == out


def test_sample_discrete():
    parameters = {'a': (1, 2), 'b': (3., 4.), 'c': 5}
    sample = ap.sample_discrete(parameters)
    assert sample == out


def test_sample_saltelli():
    parameters = {'a': (1, 2), 'b': (3., 4.), 'c': 5}
    sample = ap.sample_saltelli(parameters, n=1, digits=2)

    problem = {
        'num_vars': 2,
        'names': ['a', 'b', 'x3'],
        'bounds': [[1., 2.],
                   [3., 4.]]
    }
    param_values = saltelli.sample(problem, 1, calc_second_order=False)

    for s1, s2 in zip(sample, param_values):
        assert s1['a'] == int(round(s2[0]))
        assert s1['b'] == round(s2[1], 2)
        assert s1['c'] == 5



"""
Agentpy Grid Module
Content: Class for discrete spatial environments
"""

# There is much room for optimization in this module. Contributions welcome! :)
# TODO __init__: Add argument connect_borders
# TODO _get_diamond: distance argument, exclude original agent, & limit dmax
# TODO Method for distance between agents
# TODO items: reverse argument

import itertools
import numpy as np
import random as rd
import collections.abc as abc
from .objects import ApEnv, Agent
from .tools import make_list, make_matrix
from .lists import AgentList


class Grid(ApEnv):
    """ Environment that contains agents with a discrete spatial topology.
    Every location consists of an :class:`AgentList` that can hold
    zero, one, or more agents.
    To add new grid environments to a model, use :func:`Model.add_grid`.
    For a continuous spatial topology, use :class:`Space`.

    This class can be used as a parent class for custom grid types.
    All agentpy model objects call the method :func:`setup` after creation,
    and can access class attributes like dictionary items.
    See :class:`Environment` for general properties of all environments.

    Arguments:
        model (Model): The model instance.
        shape (tuple of int): Size of the grid.
            The length of the tuple defines the number of dimensions,
            and the values in the tuple define the length of each dimension.
        **kwargs: Will be forwarded to :func:`Grid.setup`.

    Attributes:
        grid (list of lists): Matrix of :class:`AgentList`.
        shape (tuple of int): Length of each grid dimension.
        dim (int): Number of dimensions.
    """

    def __init__(self, model, shape, **kwargs):

        super().__init__(model)

        self._topology = 'grid'
        self._grid = make_matrix(shape, AgentList)
        self._agent_dict = {}
        self._shape = shape
        self._set_var_ignore()
        self.setup(**kwargs)

    @property
    def grid(self):
        return self._grid

    @property
    def shape(self):
        return self._shape

    @property
    def dim(self):
        return len(self._shape)

    def add_agents(self, agents=1, agent_class=Agent, positions=None,  # noqa
                   random=False, generator=None, **kwargs):
        """ Adds agents to the grid environment, and returns new agents.
        See :func:`Environment.add_agents` for standard arguments.
        Additional arguments are listed below.

        Arguments:
            positions(list of tuple, optional): The positions of the added
                agents. List must have the same length as number of agents
                to be added, and each entry must be a tuple with coordinates.
                If none is passed, agents will fill up the grid systematically.
            random(bool, optional):
                If no positions are passed, agents will be placed in random
                locations instead of systematic filling (default False).
            generator (numpy.random.Generator, optional):
                Random number generator that is used if 'random' is True.
                If none is passed, :obj:`Model.random` is used.
        """

        # Standard adding
        new_agents = super().add_agents(agents, agent_class, **kwargs)

        # Extra grid features
        if not positions:
            n_agents = len(new_agents)
            available_positions = list(self.positions())

            # Extend positions if necessary
            while n_agents > len(available_positions):
                available_positions.extend(available_positions)

            # Fill agent positions
            positions = []
            if random:
                generator = generator if generator else self.model.random
                positions.extend(generator.choice(available_positions,
                                                  n_agents, replace=False))
            else:
                positions.extend(available_positions[:n_agents])

        for agent, pos in zip(new_agents, positions):
            pos = tuple(pos)  # Ensure that position is tuple
            self._agent_dict[agent] = pos  # Add position to agent_dict
            self._get_agents_from_pos(pos).append(agent)  # Add agent to pos

        return new_agents

    def remove_agents(self, agents):
        """ Removes agents from the environment. """
        for agent in make_list(agents):
            self._get_agents_from_pos(self.position(agent)).remove(agent)
            del self._agent_dict[agent]
        super().remove_agents(agents)

    @staticmethod
    def _apply(grid, func, *args, **kwargs):
        if not isinstance(grid[0], AgentList):
            return [Grid._apply(subgrid, func, *args, **kwargs)
                    for subgrid in grid]
        else:
            return [func(i, *args, **kwargs) for i in grid]

    def apply(self, func, *args, **kwargs):
        """ Applies a function to all grid positions,
        and returns grid with return values. """
        return self._apply(self.grid, func, *args, **kwargs)

    @staticmethod
    def _get_attr(agent_list, attr_key, sum_values, fill_empty):
        if len(agent_list) == 0:
            return fill_empty
        if sum_values:
            return sum(getattr(agent_list, attr_key))
        else:
            return list(getattr(agent_list, attr_key))

    def attribute(self, attr_key, sum_values=True, empty=np.nan):
        """ Returns a grid with the value of the attributes of the agents
        in each position.

        Arguments:
            attr_key(str): Name of the attribute.
            sum_values(str, optional): What to return in a position where there
                are multiple agents. If True (default), the sum of attributes.
                If False, a list of attributes.
            empty(optional): What to return for empty positions
                without agents (default numpy.nan).
        """
        return self.apply(self._get_attr, attr_key, sum_values, empty)

    def position(self, agent):
        """ Returns tuple with position of a passed agent.

        Arguments:
            agent(int or Agent): Id or instance of the agent.
        """
        if isinstance(agent, int):
            agent = self.model.get_obj(agent)
        return self._agent_dict[agent]

    def positions(self, area=None):
        """Returns iterable of all grid positions in area.

        Arguments:
            area(list of tuples, optional):
                Area of positions that should be returned.
                If none is passed, the whole grid is selected.
                Style: `[(x_start, x_end), (y_start, y_end), ...]`
        """
        if area is None:
            return itertools.product(*[range(x) for x in self._shape])
        else:
            return itertools.product(*[range(x, y+1) for x, y in area])

    def _get_agents_from_pos_rec(self, grid, pos):
        """ Recursive function to get position. """
        if len(pos) == 1:
            return grid[pos[0]]
        return self._get_agents_from_pos_rec(grid[pos[0]], pos[1:])

    def _get_agents_from_pos(self, pos):
        """ Return content of a position. """
        return self._get_agents_from_pos_rec(self._grid, pos)

    def _get_agents_from_area(self, area, grid):
        """ Recursive function to get agents in area. """
        subgrid = grid[area[0][0]:area[0][1] + 1]
        # Detect last row (must have AgentLists)
        if isinstance(subgrid[0], AgentList):
            # Flatten list of AgentLists to list of agents
            return [y for x in subgrid for y in x]
        objects = []
        for row in subgrid:
            objects.extend(self._get_agents_from_area(area[1:], row))
        return objects

    def get_agents(self, area=None):
        """ Returns an :class:`AgentList` with agents
        in the selected positions or area.

        Arguments:
            area(tuple of integers or tuples):
                Area from which agents should be gathered.
                Can either indicate a single position `[x, y, ...]`
                or an area `[(x_start, x_end), (y_start, y_end), ...]`.
        """
        if area is None:
            return self.agents
        if not isinstance(area, abc.Iterable):
            raise ValueError(f"area '{area}' is not iterable.")
        if isinstance(area[0], int):  # Soft copy
            return AgentList(self._get_agents_from_pos(area), model=self.model)
        else:
            return AgentList(
                self._get_agents_from_area(area, self._grid), model=self.model)

    def move_agent(self, agent, position):
        """ Moves agent to new position.

        Arguments:
            agent(int or Agent): Id or instance of the agent.
            position(list of int): New position of the agent.
        """
        if isinstance(agent, int):
            agent = self.model.get_obj(agent)
        self._get_agents_from_pos(self._agent_dict[agent]).remove(agent)
        self._get_agents_from_pos(position).append(agent)
        self._agent_dict[agent] = position  # Document new position

    def items(self, area=None):
        """ Returns iterator with tuples of style: (position, agents). """
        p_it_1, p_it_2 = itertools.tee(self.positions(area))  # Copy iterator
        return zip(p_it_1, [self._get_agents_from_pos(pos)
                            for pos in p_it_2])

    def _get_diamond(self, pos, grid, dist=1):
        """ Return agents in diamond-shaped area around pos."""
        subgrid = grid[max(0, pos[0] - dist):pos[0] + dist + 1]
        if len(pos) == 1:
            return [y for x in subgrid for y in x]  # flatten list
        objects = []
        for row, dist in zip(subgrid, [0, 1, 0]):
            objects.extend(self._get_diamond(pos[1:], row, dist))
        return objects

    def neighbors(self, agent, distance=1, diagonal=True):
        """ Returns agent neighbors.

        Arguments:
            agent(int or Agent): Id or instance of the agent.
            distance(int, optional): Number of positions to cover in each
                direction.
            diagonal(bool, optional):
                If True (default), diagonal neighbors are included.
                If False, only direct neighbors are included
                (currently only works with distance == 1).
        """
        if isinstance(agent, int):
            agent = self.model.get_obj(agent)
        if diagonal:  # Include diagonal neighbors
            a_pos = self._agent_dict[agent]
            area = [(max(p-distance, 0),
                     min(p+distance, dmax))
                    for p, dmax in zip(a_pos, self._shape)]
            agents = self.get_agents(area)
        else:  # Do not include diagonal neighbors
            agents = self._get_diamond(self._agent_dict[agent], self._grid)
        agents.remove(agent)  # Remove original agent
        return AgentList(agents, model=self.model)



import pytest
import agentpy as ap
import numpy as np
import matplotlib.pyplot as plt


def test_gridplot():
    """Test only for errors."""

    # Use cmap values
    grid1 = [[1, 1], [1, np.nan]]
    ap.gridplot(grid1)

    # Use RGB values
    x = (1., 1., 0.)
    grid2 = [[x, x], [x, x]]
    ap.gridplot(grid2)

    # Use color dict and convert
    cdict = {1: 'g'}
    ap.gridplot(grid1, color_dict=cdict, convert=True)

    # Use only convert
    grid3 = [['g', 'g'], ['g', np.nan]]
    ap.gridplot(grid3, convert=True)

    # Assign to axis
    fig, ax = plt.subplots()
    ap.gridplot(grid1, ax=ax)

    assert True


def test_animation():
    """Test only for errors."""
    def my_plot(model, ax):
        ax.set_title(f"{model.t}")

    fig, ax = plt.subplots()
    my_model = ap.Model({'steps': 2})
    animation = ap.animate(my_model, fig, ax, my_plot, skip=1)
    animation.to_jshtml()

    # Stop immediately
    my_model = ap.Model({'steps': 0})
    animation = ap.animate(my_model, fig, ax, my_plot)
    animation.to_jshtml()

    # Skip more than steps
    my_model = ap.Model({'steps': 0})
    animation = ap.animate(my_model, fig, ax, my_plot, skip=1)
    animation.to_jshtml()

    assert True


class MyModel(ap.Model):
    def step(self):
        self.measure('x', self.p.x)
        self.stop()


def test_sobol():
    si = 0.6593259637723373

    param_ranges = {'x': (0., 1.)}
    sample = ap.sample_saltelli(param_ranges, n=10)
    results = ap.Experiment(MyModel, sample).run()
    ap.sensitivity_sobol(results, param_ranges, measures='x')
    assert results.sensitivity['S1'][0] == si

    # Test if a non-varied parameter causes errors
    param_ranges = {'x': (0., 1.), 'y': 1}
    sample = ap.sample_saltelli(param_ranges, n=10)
    results = ap.Experiment(MyModel, sample).run()
    ap.sensitivity_sobol(results, param_ranges)
    assert results.sensitivity['S1'][0] == si

    # Test calc_second_order
    param_ranges = {'x': (0., 1.), 'y': 1}
    sample = ap.sample_saltelli(param_ranges, n=10, calc_second_order=True)
    results = ap.Experiment(MyModel, sample).run()
    ap.sensitivity_sobol(results, param_ranges, calc_second_order=True)
    assert results.sensitivity[('S2', 'x')][0].__repr__() == 'nan'



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="agentpy",
    license="BSD 3-Clause",
    author="Joël Foramitti",
    author_email="joel.foramitti@uab.cat",
    description="Agent-based modeling in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://agentpy.readthedocs.io/",
    download_url="https://github.com/JoelForamitti/agentpy",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent"
    ],
)



"""
Agentpy Model Module
Content: Main class for agent-based models
"""

import numpy as np
import pandas as pd
from datetime import datetime

from .output import DataDict
from .objects import ApEnv, Agent, Environment
from .network import Network
from .grid import Grid
from .space import Space
from .tools import AttrDict, AgentpyError, make_list
from .lists import ObjList


class Model(ApEnv):
    """
    An agent-based model that can hold environments and agents.

    This class can be used as a parent class for custom models.
    Class attributes can be accessed like dictionary items.
    To define the procedures of a simulation, override the methods
    :func:`Model.setup`, :func:`Model.step`,
    :func:`Model.update`, and :func:`Model.end`.
    See :func:`Model.run` for more information on the simulation procedure.

    Attributes:
        name (str): The models' name.
        envs (EnvList): The models' environments.
        agents (AgentList): The models' agents.
        p (AttrDict): The models' parameters.
        t (int): Current time-step of the model.
        log (dict): The models' recorded variables.
        output (DataDict): Output data after simulation.

    Arguments:
        parameters (dict, optional): Dictionary of model parameters.
            Recommended types for parameters are int, float, str, list,
            numpy.integer, numpy.floating, and numpy.ndarray.
            Other types might cause errors.
        run_id (int, optional): Number of current run (default None).
        scenario (str, optional): Current scenario (default None).
        **kwargs: Will be forwarded to :func:`Model.setup`
    """

    def __init__(self, parameters=None, run_id=None, scenario=None, **kwargs):

        self._id_counter = -1
        self._obj_dict = {}  # Objects mapped by their id
        super().__init__(self)  # Model will be have id 0

        self.t = 0
        self.run_id = run_id
        self.scenario = scenario

        # Recording
        self._measure_log = {}
        self.output = DataDict()
        self.output.log = {'model_type': self.type,
                           'time_stamp': str(datetime.now())}

        # Private variables
        self._random = np.random.default_rng()
        self._steps = None
        self._parameters = AttrDict(parameters)
        self._stop = False
        self._set_var_ignore()
        self._setup_kwargs = kwargs

    def __repr__(self):
        rep = f"Agent-based model {{"
        keys = ['type', 'agents', 'envs', 'p']
        items = [(k, self[k]) for k in keys]
        items += list(self.__dict__.items())
        for k, v in items:
            if k[0] != '_':
                v = v._short_repr() if '_short_repr' in dir(v) else v
                rep += f"\n'{k}': {v}"
        return rep + '\n}'

    @property
    def objects(self):
        """ Returns a list of all model objects (agents and environments). """
        return ObjList(self.agents + self.envs)

    @property
    def random(self):
        """ Returns the models random number generator
        of type :class:`numpy.random.Generator`. """
        return self._random

    def get_obj(self, obj_id):
        """ Returns model object with passed object id (int). """
        try:
            return self._obj_dict[obj_id]
        except KeyError:
            raise ValueError(f"Model has no object with obj_id '{obj_id}'.")

    def _new_id(self):
        """ Returns a new unique object id (int). """
        self._id_counter += 1
        return self._id_counter

    def add_env(self, env_class=Environment, **kwargs):
        """ Creates a new environment. """
        new_env = env_class(self.model, **kwargs)
        self.envs.append(new_env)
        return new_env

    def add_network(self, graph=None, agents=None, **kwargs):
        """ Creates a new environment with a network.
        Arguments are forwarded to :class:`Network`. """
        new_env = Network(self.model, graph=graph, agents=agents, **kwargs)
        self.envs.append(new_env)
        return new_env

    def add_grid(self, shape, **kwargs):
        """ Creates a new environment with a spatial grid.
        Arguments are forwarded to :class:`Grid`. """
        new_env = Grid(self.model, shape=shape, **kwargs)
        self.envs.append(new_env)
        return new_env

    def add_space(self, shape, **kwargs):
        """ Creates a new environment with a continuous space.
        Arguments are forwarded to :class:`Space`. """
        new_env = Space(self.model, shape=shape, **kwargs)
        self.envs.append(new_env)
        return new_env

    def measure(self, measure, value):
        """ Records an evaluation measure. """
        self._measure_log[measure] = [value]

    # Main simulation functions

    def setup(self, **kwargs):
        """ Defines the model's actions before the first simulation step.
        Can be overwritten and used to initiate agents and environments."""
        pass

    def step(self):
        """ Defines the model's actions during each simulation step.
        Can be overwritten and used to set the models' main dynamics."""
        pass

    def update(self):
        """ Defines the model's actions after setup and each simulation step.
        Can be overwritten and used for the recording of dynamic variables. """
        pass

    def end(self):
        """ Defines the model's actions after the last simulation step.
        Can be overwritten and used for final calculations and measures."""
        pass

    def stop(self):
        """ Stops :meth:`Model.run` during an active simulation. """
        self._stop = True

    def _setup_run(self, steps=None, seed=None):
        """ Prepare round 0 of a simulation.
        See Model.run() for more informatin. """

        # Prepare random generator
        if not seed and 'seed' in self.p:
            seed = self.p['seed']  # Take seed from parameters
        if seed:
            self._random = np.random.default_rng(seed=seed)

        # Prepare steps
        if steps is None:
            self._steps = self.p['steps'] if 'steps' in self.p else 1000
        else:
            self._steps = steps

        # Initiate simulation
        self._stop = False

        # Execute setup and first update
        self.setup(**self._setup_kwargs)
        self.update()

        # Stop simulation if t too high
        if self.t >= self._steps:
            self._stop = True

    def _make_step(self):
        """ Proceed simulation by one step. """
        self.t += 1
        self.step()
        self.update()
        if self.t >= self._steps:
            self._stop = True

    def run(self, steps=None, seed=None, display=True):
        """ Executes the simulation of the model.

        The simulation proceeds as follows.
        It starts by calling :func:`Model.setup` and :func:`Model.update`.
        After that, :attr:`Model.t` is increased by 1 and
        :func:`Model.step` and :func:`Model.update` are called.
        This step is repeated until the method :func:`Model.stop` is called
        or steps is reached. After the last step, :func:`Model.end` is called.

        Arguments:
            steps (int, optional):
                Maximum number of steps for the simulation to run.
                If none is given, the parameter 'Model.p.steps' will be used.
                If there is no such parameter, 'steps' will be set to 1000.
            seed (int, optional):
                Seed to set for :obj:`Model.random` at the beginning of the simulation.
                If none is given, the parameter 'Model.p.seed' will be used.
                If there is no such parameter, as random seed will be set.
            display (bool, optional):
                Whether to display simulation progress (default True).

        Returns:
            DataDict: Recorded model data,
            which can also be found in :attr:`Model.output`.
        """

        dt0 = datetime.now()  # Time-Stamp
        self._setup_run(steps, seed)

        while not self._stop:
            self._make_step()
            if display:
                print(f"\rCompleted: {self.t} steps", end='')

        self.end()
        self._create_output()
        self.output.log['run_time'] = ct = str(datetime.now() - dt0)
        self.output.log['steps'] = self.t

        if display:
            print(f"\nRun time: {ct}\nSimulation finished")

        return self.output

    def _create_output(self):
        """ Generates an 'output' dictionary out of object logs. """

        def output_from_obj_list(self, obj_list, columns):
            # Aggregate logs per object type
            obj_types = {}
            for obj in obj_list:

                if obj.log:  # Check for variables

                    # Add object id/key to object log
                    obj.log['obj_id'] = [obj.id] * len(obj.log['t'])

                    # Initiate object type if new
                    obj_type = type(obj).__name__

                    if obj_type not in obj_types.keys():
                        obj_types[obj_type] = {}

                    # Add object log to aggr. log
                    for k, v in obj.log.items():
                        if k not in obj_types[obj_type]:
                            obj_types[obj.type][k] = []
                        obj_types[obj_type][k].extend(v)

            # Transform logs into dataframes
            for obj_type, log in obj_types.items():
                df = pd.DataFrame(log)
                for k, v in columns.items():
                    df[k] = v  # Set additional index columns
                df = df.set_index(list(columns.keys()) + ['obj_id', 't'])
                self.output['variables'][obj_type] = df

        # 0 - Document parameters
        if self.p:
            self.output['parameters'] = self.p

        # 1 - Define additional index columns
        columns = {}
        if self.run_id is not None:
            columns['run_id'] = self.run_id
        if self.scenario is not None:
            columns['scenario'] = self.scenario

        # 2 - Create measure output
        if self._measure_log:
            d = self._measure_log
            for key, value in columns.items():
                d[key] = value
            df = pd.DataFrame(d)
            if columns:
                df = df.set_index(list(columns.keys()))
            self.output['measures'] = df

        # 3 - Create variable output
        self.output['variables'] = DataDict()

        # 3.1 - Create variable output for objects
        output_from_obj_list(self, self.agents, columns)
        output_from_obj_list(self, self.envs, columns)

        # 3.2 - Create variable output for model
        if self.log:
            df = pd.DataFrame(self.log)
            # df['obj_id'] = 'model'
            for k, v in columns.items():
                df[k] = v
            df = df.set_index(list(columns.keys()) + ['t'])  # 'obj_id',

            if self.output['variables']:
                self.output['variables'][self.type] = df
            else:
                self.output['variables'] = df  # No subdict if only model vars

        # 3.3 - Remove variable dict if empty (i.e. nothing has been added)
        elif not self.output['variables']:
            del self.output['variables']



import pytest
import networkx as nx
import agentpy as ap


def test_add_agents():

    # Add agents to existing nodes
    graph = nx.Graph()
    graph.add_node(0)
    graph.add_node(1)
    model = ap.Model()
    model.add_agents(2)
    model.add_network(graph=graph, agents=model.agents)
    model.env.graph.add_edge(model.agents[0],model.agents[1])
    assert list(model.agents[0].neighbors().id) == [2]

    # Add agents as new nodes
    model2 = ap.Model()
    agents = model2.add_agents(2)
    model2.add_network(agents = agents[0])  # Add at initialization
    model2.env.add_agents(agents[1])  # Add later
    model2.env.graph.add_edge(model2.agents[0],model2.agents[1])

    # Test if the two graphs are identical
    assert model.env.graph.nodes.__repr__() == model2.env.graph.nodes.__repr__()
    assert model.env.graph.edges.__repr__() == model2.env.graph.edges.__repr__()

    # Test errors
    model3 = ap.Model()
    graph = nx.Graph()
    model3.add_agents()
    with pytest.raises(ValueError):
        assert model3.add_network(graph=graph, agents=model3.agents)
    with pytest.raises(TypeError):
        assert model3.add_network(graph=1)


def test_remove_agents():

    model = ap.Model()
    model.add_agents(2)
    nw = model.add_network()
    nw.add_agents(model.agents)
    agent = model.agents[0]
    nw.remove_agents(agent)
    len(nw.agents) == 1
    len(nw.graph.nodes) == 1


"""
Agentpy Analysis Module
Content: Sensitivity and interactive analysis, animation, visualization
"""

import matplotlib
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from SALib.analyze import sobol

from .tools import make_list, param_tuples_to_salib


def _sobol_set_df_index(df, p_keys, measure):
    df['parameter'] = p_keys
    df['measure'] = measure
    df.set_index(['measure', 'parameter'], inplace=True)


def sensitivity_sobol(output, param_ranges, measures=None,
                      calc_second_order=False, **kwargs):
    """ Calculates Sobol Sensitivity Indices and adds them to the output,
    using :func:`SALib.analyze.sobol.analyze`. To work correctly,
    data must be from an experiment with a parameter sample that was
    generated by :func:`sample_saltelli`.

    Arguments:
        output (DataDict): The output of an experiment that was set to only
            one iteration (default) and used a parameter sample that was
            generated with :func:`sample_saltelli`.
        param_ranges (dict): The same dictionary that was used for the
            generation of the parameter sample with :func:`sample_saltelli`.
        measures (str or list of str, optional): The measures that should
            be used for the analysis. If none are passed, all are used.
        calc_second_order (bool, optional): Whether to calculate second order
            indices (default False). Value must be the same as the one used
            to generate the parameter sample with :func:`sample_saltelli`.
        **kwargs: Will be forwarded to :func:`SALib.analyze.sobol.analyze`.
    """

    # STEP 1 - Convert param_ranges to SALib Format
    param_ranges_tuples = {k: v for k, v in param_ranges.items()
                           if isinstance(v, tuple)}
    param_ranges_salib = param_tuples_to_salib(param_ranges_tuples)

    # STEP 2 - Calculate Sobol Sensitivity Indices
    if measures is None:
        measures = output.measures.columns
    if isinstance(measures, str):
        measures = [measures]
    p_keys = output._combine_pars(varied=True, fixed=False).keys()
    dfs_list = [[] for _ in range(4 if calc_second_order else 2)]

    for measure in measures:
        y = np.array(output.measures[measure])
        si = sobol.analyze(param_ranges_salib, y, calc_second_order, **kwargs)

        # Make dataframes out of S1 and ST sensitivities
        keyss = [['S1', 'ST'], ['S1_conf', 'ST_conf']]
        for keys, dfs in zip(keyss, dfs_list[0:2]):
            s = {k[0:2]: v for k, v in si.items() if k in keys}
            df = pd.DataFrame(s)
            _sobol_set_df_index(df, p_keys, measure)
            dfs.append(df)

        # Make dataframes out S2 sensitivities
        if calc_second_order:
            for key, dfs in zip(['S2', 'S2_conf'], dfs_list[2:4]):
                df = pd.DataFrame(si[key])
                _sobol_set_df_index(df, p_keys, measure)
                dfs.append(df)

    # Combine dataframes for each measure
    output['sensitivity'] = pd.concat(dfs_list[0])
    output['sensitivity_conf'] = pd.concat(dfs_list[1])

    if calc_second_order:

        # Add Second-Order to Output
        dfs_si = [output['sensitivity'], pd.concat(dfs_list[2])]
        dfs_si_conf = [output['sensitivity_conf'], pd.concat(dfs_list[3])]
        output['sensitivity'] = pd.concat(dfs_si, axis=1)
        output['sensitivity_conf'] = pd.concat(dfs_si_conf, axis=1)

        # Create Multi-Index for Columns
        arrays = [["S1", "ST"] + ["S2"] * len(p_keys), [""] * 2 + list(p_keys)]
        tuples = list(zip(*arrays))
        index = pd.MultiIndex.from_tuples(tuples, names=["order", "parameter"])
        output['sensitivity'].columns = index
        output['sensitivity_conf'].columns = index.copy()

    return output


def animate(model, fig, axs, plot, steps=None, seed=None,
            skip=0, fargs=(), **kwargs):
    """ Returns an animation of the model simulation,
    using :func:`matplotlib.animation.FuncAnimation`.

    Arguments:
        model (Model): The model instance.
        fig (matplotlib.figure.Figure): Figure for the animation.
        axs (matplotlib.axes.Axes or list): Axis or list of axis of the figure.
        plot (function): Function that takes `(model, ax, *fargs)`
            and creates the desired plots on each axis at each time-step.
        steps(int, optional):
            Maximum number of steps for the simulation to run.
            If none is given, the parameter 'Model.p.steps' will be used.
            If there is no such parameter, 'steps' will be set to 1000.
        seed (int, optional):
            Seed to set for :obj:`Model.random` at the beginning of the simulation.
            If none is given, the parameter 'Model.p.seed' will be used.
            If there is no such parameter, as random seed will be set.
        skip (int, optional): Number of rounds to skip before the
            animation starts (default 0).
        fargs (tuple, optional): Forwarded fo the `plot` function.
        **kwargs: Forwarded to :func:`matplotlib.animation.FuncAnimation`.

    Examples:
        An animation can be generated as follows::

            def my_plot(model, ax):
                pass  # Call pyplot functions here
            
            fig, ax = plt.subplots() 
            my_model = MyModel(parameters)
            animation = ap.animate(my_model, fig, ax, my_plot)

        One way to display the resulting animation object in Jupyter::

            from IPython.display import HTML
            HTML(animation.to_jshtml())
    """

    model._setup_run(steps, seed)
    model._create_output()
    pre_steps = 0

    for _ in range(skip):
        model._make_step()

    def frames():
        nonlocal model, pre_steps
        if model._stop is False:
            while not model._stop:
                if pre_steps < 2:  # Frames iterates twice before starting plot
                    pre_steps += 1
                else:
                    model._make_step()
                    model._create_output()
                yield model.t
        else:  # Yield current if model stops before the animation starts
            yield model.t

    def update(t, m, axs, *fargs):  # noqa
        nonlocal pre_steps
        for ax in make_list(axs):
            # Clear axes before each plot
            ax.clear()
        plot(m, axs, *fargs)  # Perform plot

    ani = matplotlib.animation.FuncAnimation(
        fig, update,
        frames=frames,
        fargs=(model, axs, *fargs),
        save_count=model._steps,
        **kwargs)  # noqa

    plt.close()  # Don't display static plot
    return ani


def _apply_colors(grid, color_dict, convert):
    if isinstance(grid[0], list):
        return [_apply_colors(subgrid, color_dict, convert)
                for subgrid in grid]
    else:
        if color_dict is not None:
            grid = [i if i is np.nan else color_dict[i] for i in grid]
        if convert is True:
            grid = [(0., 0., 0., 0.) if i is np.nan else
                    matplotlib.colors.to_rgba(i) for i in grid]
        return grid


def gridplot(grid, color_dict=None, convert=False, ax=None, **kwargs):
    """ Visualizes values on a two-dimensional grid with
    :func:`matplotlib.pyplot.imshow`.

    Arguments:
        grid(list of list): Two-dimensional grid with values.
            numpy.nan values will be plotted as empty patches.
        color_dict(dict, optional): Dictionary that translates
            each value in `grid` to a color specification.
        convert(bool, optional): Convert values to rgba vectors,
             using :func:`matplotlib.colors.to_rgba` (default False).
        ax(matplotlib.pyplot.axis, optional): Axis to be used for plot.
        **kwargs: Forwarded to :func:`matplotlib.pyplot.imshow`.
     """

    # TODO Make feature for legend
    if color_dict is not None or convert:
        grid = _apply_colors(grid, color_dict, convert)
    if ax:
        ax.imshow(grid, **kwargs)
    else:
        plt.imshow(grid, **kwargs)
